import numpy as np
import pytest
from astropy.io import fits
from stcal.jump.jump import (
    calc_num_slices,
    extend_saturation,
    find_ellipses,
    find_faint_extended,
    flag_large_events,
    point_inside_ellipse,
    find_first_good_group,
    detect_jumps,
    find_last_grp
)

DQFLAGS = {"JUMP_DET": 4, "SATURATED": 2, "DO_NOT_USE": 1, "GOOD": 0, "NO_GAIN_VALUE": 8,
           "REFERENCE_PIXEL": 2147483648}


@pytest.fixture()
def setup_cube():
    def _cube(ngroups, readnoise=10):
        nints = 1
        nrows = 204
        ncols = 204
        rej_threshold = 3
        nframes = 1
        data = np.zeros(shape=(nints, ngroups, nrows, ncols), dtype=np.float32)
        read_noise = np.full((nrows, ncols), readnoise, dtype=np.float32)
        gdq = np.zeros(shape=(nints, ngroups, nrows, ncols), dtype=np.uint32)

        return data, gdq, nframes, read_noise, rej_threshold

    return _cube

def test_nirspec_saturated_pix():
    """
    This test is based on an actual NIRSpec exposure that has some pixels
    flagged as saturated in one or more groups, which the jump step is
    supposed to ignore, but an old version of the code was setting JUMP flags
    for some of the saturated groups. This is to verify that the saturated
    groups are no longer flagged with jumps.
    """
    ingain = 1.0
    inreadnoise = 10.7
    ngroups = 7
    nrows = 2
    ncols = 2
    nints = 1
    nframes = 1
    data = np.zeros(shape=(nints, ngroups, nrows, ncols), dtype=np.float32)
    read_noise = np.full((nrows, ncols), inreadnoise, dtype=np.float32)
    gdq = np.zeros(shape=(nints, ngroups, nrows, ncols), dtype=np.uint32)
    err = np.zeros(shape=(nrows, ncols), dtype=np.float32)
    pdq = np.zeros(shape=(nrows, ncols), dtype=np.uint32)
    gain = np.ones_like(read_noise) * ingain

    # Setup the needed input pixel and DQ values
    data[0, :, 1, 1] = [639854.75, 4872.451, -17861.791, 14022.15, 22320.176,
                              1116.3828, 1936.9746]
    gdq[0, :, 1, 1] = [0, 0, 0, 0, 0, 2, 2]
    data[0, :, 0, 1] = [8.25666812e+05, -1.10471914e+05, 1.95755371e+02, 1.83118457e+03,
                              1.72250879e+03, 1.81733496e+03, 1.65188281e+03]
    # 2 non-sat groups means only 1 non-sat diff, so no jumps should be flagged
    gdq[0, :, 0, 1] = [0, 0, 2, 2, 2, 2, 2]
    data[0, :, 1, 0] = [1228767., 46392.234, -3245.6553, 7762.413,
                              37190.76, 266611.62, 5072.4434]
    gdq[0, :, 1, 0] = [0, 0, 0, 0, 0, 0, 2]

    # run jump detection
    gdq, pdq, total_primary_crs, number_extended_events, stddev = detect_jumps(nframes, data, gdq, pdq, err,
                                                                               gain, read_noise, rejection_thresh=4.0,
                                                                               three_grp_thresh=5,
                                 four_grp_thresh=6,
                                 max_cores='none', max_jump_to_flag_neighbors=200,
                                 min_jump_to_flag_neighbors=10, flag_4_neighbors=True, dqflags=DQFLAGS)

    # Check the results. There should not be any pixels with DQ values of 6, which
    # is saturated (2) plus jump (4). All the DQ's should be either just 2 or just 4.
    np.testing.assert_array_equal(gdq[0, :, 1, 1], [0, 4, 0, 4, 4, 2, 2])
    # assert that no groups are flagged when there's only 1 non-sat. grp
    np.testing.assert_array_equal(gdq[0, :, 0, 1], [0, 0, 2, 2, 2, 2, 2])
    np.testing.assert_array_equal(gdq[0, :, 1, 0], [0, 4, 4, 0, 4, 4, 2])

def test_multiprocessing():
    nints = 1
    nrows = 13
    ncols = 2
    ngroups = 13
    readnoise = 10
    frames_per_group = 1

    data = np.zeros(shape=(nints, ngroups, nrows, ncols), dtype=np.float32)
    readnoise_2d = np.ones((nrows, ncols), dtype=np.float32) * readnoise
    gain_2d = np.ones((nrows, ncols), dtype=np.float32) * 4
    gdq = np.zeros(shape=(nints, ngroups, nrows, ncols), dtype=np.uint32)
    pdq = np.zeros(shape=(nrows, ncols), dtype=np.uint32)
    err = np.zeros(shape=(nrows, ncols), dtype=np.float32)
    num_cores = "1"
    data[0, 4:, 5, 1] = 2000
    gdq[0, 4:, 6, 1] = DQFLAGS['DO_NOT_USE']
    gdq, pdq, total_primary_crs, number_extended_events, stddev = detect_jumps(
        frames_per_group, data, gdq, pdq, err, gain_2d, readnoise_2d, rejection_thresh=5, three_grp_thresh=6,
        four_grp_thresh=7, max_cores=num_cores, max_jump_to_flag_neighbors=10000, min_jump_to_flag_neighbors=100,
        flag_4_neighbors=True, dqflags=DQFLAGS)
    print(data[0, 4, :, :])
    print(gdq[0, 4, :, :])
    assert gdq[0, 4, 5, 1] == DQFLAGS['JUMP_DET']
    assert gdq[0, 4, 6, 1] == DQFLAGS['DO_NOT_USE']

    # This section of code will fail without the fixes for PR #239 that prevent
    # the double flagging pixels with jump which already have do_not_use or saturation set.
    num_cores = "5"
    data = np.zeros(shape=(nints, ngroups, nrows, ncols), dtype=np.float32)
    gdq = np.zeros(shape=(nints, ngroups, nrows, ncols), dtype=np.uint32)
    pdq = np.zeros(shape=(nrows, ncols), dtype=np.uint32)
    readnoise_2d = np.ones((nrows, ncols), dtype=np.float32) * readnoise
    gain_2d = np.ones((nrows, ncols), dtype=np.float32) * 3
    err = np.zeros(shape=(nrows, ncols), dtype=np.float32)
    data[0, 4:, 5, 1] = 2000
    gdq[0, 4:, 6, 1] = DQFLAGS['DO_NOT_USE']
    gdq, pdq, total_primary_crs, number_extended_events, stddev = detect_jumps(
        frames_per_group, data, gdq, pdq, err, gain_2d, readnoise_2d, rejection_thresh=5, three_grp_thresh=6,
        four_grp_thresh=7, max_cores=num_cores, max_jump_to_flag_neighbors=10000, min_jump_to_flag_neighbors=100,
        flag_4_neighbors=True, dqflags=DQFLAGS)
    assert gdq[0, 4, 5, 1] == DQFLAGS['JUMP_DET']
    assert gdq[0, 4, 6, 1] == DQFLAGS['DO_NOT_USE'] #This value would have been 5 without the fix.


def test_multiprocessing_big():
    nints = 1
    nrows = 2048
    ncols = 7
    ngroups = 13
    readnoise = 10
    frames_per_group = 1

    data = np.zeros(shape=(nints, ngroups, nrows, ncols), dtype=np.float32)
    readnoise_2d = np.ones((nrows, ncols), dtype=np.float32) * readnoise
    gain_2d = np.ones((nrows, ncols), dtype=np.float32) * 4
    gdq = np.zeros(shape=(nints, ngroups, nrows, ncols), dtype=np.uint32)
    pdq = np.zeros(shape=(nrows, ncols), dtype=np.uint32)
    err = np.zeros(shape=(nrows, ncols), dtype=np.float32)
    num_cores = "1"
    data[0, 4:, 204, 5] = 2000
    gdq[0, 4:, 204, 6] = DQFLAGS['DO_NOT_USE']
    gdq, pdq, total_primary_crs, number_extended_events, stddev = detect_jumps(
        frames_per_group, data, gdq, pdq, err, gain_2d, readnoise_2d, rejection_thresh=5, three_grp_thresh=6,
        four_grp_thresh=7, max_cores=num_cores, max_jump_to_flag_neighbors=10000, min_jump_to_flag_neighbors=100,
        flag_4_neighbors=True, dqflags=DQFLAGS)
    print(data[0, 4, :, :])
    print(gdq[0, 4, :, :])
    assert gdq[0, 4, 204, 5] == DQFLAGS['JUMP_DET']
    assert gdq[0, 4, 205, 5] == DQFLAGS['JUMP_DET']
    assert gdq[0, 4, 204, 6] == DQFLAGS['DO_NOT_USE']

    # This section of code will fail without the fixes for PR #239 that prevent
    # the double flagging pixels with jump which already have do_not_use or saturation set.
    num_cores = "10"
    data = np.zeros(shape=(nints, ngroups, nrows, ncols), dtype=np.float32)
    gdq = np.zeros(shape=(nints, ngroups, nrows, ncols), dtype=np.uint32)
    pdq = np.zeros(shape=(nrows, ncols), dtype=np.uint32)
    readnoise_2d = np.ones((nrows, ncols), dtype=np.float32) * readnoise
    gain_2d = np.ones((nrows, ncols), dtype=np.float32) * 3
    err = np.zeros(shape=(nrows, ncols), dtype=np.float32)
    data[0, 4:, 204, 5] = 2000
    gdq[0, 4:, 204, 6] = DQFLAGS['DO_NOT_USE']
    gdq, pdq, total_primary_crs, number_extended_events, stddev = detect_jumps(
        frames_per_group, data, gdq, pdq, err, gain_2d, readnoise_2d, rejection_thresh=5, three_grp_thresh=6,
        four_grp_thresh=7, max_cores=num_cores, max_jump_to_flag_neighbors=10000, min_jump_to_flag_neighbors=100,
        flag_4_neighbors=True, dqflags=DQFLAGS)
    assert gdq[0, 4, 204, 5] == DQFLAGS['JUMP_DET']
    assert gdq[0, 4, 205, 5] == DQFLAGS['JUMP_DET']
    assert gdq[0, 4, 204, 6] == DQFLAGS['DO_NOT_USE'] #This value would have been 5 without the fix.


def test_find_simple_ellipse():
    plane = np.zeros(shape=(5, 5), dtype=np.uint8)
    plane[2, 2] = DQFLAGS["JUMP_DET"]
    plane[3, 2] = DQFLAGS["JUMP_DET"]
    plane[1, 2] = DQFLAGS["JUMP_DET"]
    plane[2, 3] = DQFLAGS["JUMP_DET"]
    plane[2, 1] = DQFLAGS["JUMP_DET"]
    plane[1, 3] = DQFLAGS["JUMP_DET"]
    plane[2, 4] = DQFLAGS["JUMP_DET"]
    plane[3, 3] = DQFLAGS["JUMP_DET"]
    ellipse = find_ellipses(plane, DQFLAGS["JUMP_DET"], 1)
    assert ellipse[0][2] == pytest.approx(45.0, 1e-3)  # 90 degree rotation
    assert ellipse[0][0] == pytest.approx((2.5, 2.0))  # center


def test_find_ellipse2():
    plane = np.zeros(shape=(5, 5), dtype=np.uint8)
    plane[1, :] = [0, DQFLAGS["JUMP_DET"], DQFLAGS["JUMP_DET"], DQFLAGS["JUMP_DET"], 0]
    plane[2, :] = [0, DQFLAGS["JUMP_DET"], DQFLAGS["JUMP_DET"], DQFLAGS["JUMP_DET"], 0]
    plane[3, :] = [0, DQFLAGS["JUMP_DET"], DQFLAGS["JUMP_DET"], DQFLAGS["JUMP_DET"], 0]
    ellipses = find_ellipses(plane, DQFLAGS["JUMP_DET"], 1)
    ellipse = ellipses[0]
    assert ellipse[0][0] == 2
    assert ellipse[0][1] == 2
    assert ellipse[1][0] == 2
    assert ellipse[1][1] == 2
    assert ellipse[2] == 90.0


def test_extend_saturation_simple():
    cube = np.zeros(shape=(5, 7, 7), dtype=np.uint8)
    persist_jumps = np.zeros(shape=(7, 7), dtype=np.uint8)
    grp = 1
    min_sat_radius_extend = 1
    cube[1, 3, 3] = DQFLAGS["SATURATED"]
    cube[1, 2, 3] = DQFLAGS["SATURATED"]
    cube[1, 3, 4] = DQFLAGS["SATURATED"]
    cube[1, 4, 3] = DQFLAGS["SATURATED"]
    cube[1, 3, 2] = DQFLAGS["SATURATED"]
    cube[1, 2, 2] = DQFLAGS["JUMP_DET"]
    sat_circles = find_ellipses(cube[grp, :, :], DQFLAGS["SATURATED"], 1)
    new_cube, persist_jumps = extend_saturation(
        cube, grp, sat_circles, DQFLAGS["SATURATED"], DQFLAGS["JUMP_DET"],
        1.1, persist_jumps,
    )

    assert new_cube[grp, 2, 2] == DQFLAGS["SATURATED"]
    assert new_cube[grp, 4, 4] == DQFLAGS["SATURATED"]
    assert new_cube[grp, 4, 5] == 0


def test_flag_large_events_nosnowball():
    cube = np.zeros(shape=(1, 5, 7, 7), dtype=np.uint8)
    # cross of saturation with no jump
    cube[0, 0:2, 3, 3] = DQFLAGS["SATURATED"]
    cube[0, 0:2, 2, 3] = DQFLAGS["SATURATED"]
    cube[0, 0:2, 3, 4] = DQFLAGS["SATURATED"]
    cube[0, 0:2, 4, 3] = DQFLAGS["SATURATED"]
    cube[0, 0:2, 3, 2] = DQFLAGS["SATURATED"]
    # cross of saturation surrounding by jump -> snowball but sat core is not new
    # should have no snowball trigger
    cube[0, 2, 3, 3] = DQFLAGS["SATURATED"]
    cube[0, 2, 2, 3] = DQFLAGS["SATURATED"]
    cube[0, 2, 3, 4] = DQFLAGS["SATURATED"]
    cube[0, 2, 4, 3] = DQFLAGS["SATURATED"]
    cube[0, 2, 3, 2] = DQFLAGS["SATURATED"]
    cube[0, 2, 1, 1:6] = DQFLAGS["JUMP_DET"]
    cube[0, 2, 5, 1:6] = DQFLAGS["JUMP_DET"]
    cube[0, 2, 1:6, 1] = DQFLAGS["JUMP_DET"]
    cube[0, 2, 1:6, 5] = DQFLAGS["JUMP_DET"]
    flag_large_events(
        cube,
        DQFLAGS["JUMP_DET"],
        DQFLAGS["SATURATED"],
        min_sat_area=1,
        min_jump_area=6,
        expand_factor=1.9,
        edge_size=1,
        sat_required_snowball=True,
        min_sat_radius_extend=1,
        sat_expand=1.1,
    )
    assert cube[0, 2, 2, 2] == 0
    assert cube[0, 2, 3, 6] == 0


def test_flag_large_events_withsnowball():
    cube = np.zeros(shape=(1, 5, 7, 7), dtype=np.uint8)
    # cross of saturation surrounding by jump -> snowball
    cube[0, 2, 3, 3] = DQFLAGS["SATURATED"]
    cube[0, 2, 2, 3] = DQFLAGS["SATURATED"]
    cube[0, 2, 3, 4] = DQFLAGS["SATURATED"]
    cube[0, 2, 4, 3] = DQFLAGS["SATURATED"]
    cube[0, 2, 3, 2] = DQFLAGS["SATURATED"]
    cube[0, 2, 1, 1:6] = DQFLAGS["JUMP_DET"]
    cube[0, 2, 5, 1:6] = DQFLAGS["JUMP_DET"]
    cube[0, 2, 1:6, 1] = DQFLAGS["JUMP_DET"]
    cube[0, 2, 1:6, 5] = DQFLAGS["JUMP_DET"]
    cube, total_snowballs = flag_large_events(
        cube,
        DQFLAGS["JUMP_DET"],
        DQFLAGS["SATURATED"],
        min_sat_area=1,
        min_jump_area=6,
        expand_factor=1.9,
        edge_size=0,
        sat_required_snowball=True,
        min_sat_radius_extend=0.5,
        sat_expand=1.1,
    )
    assert cube[0, 1, 2, 2] == 0
    assert cube[0, 1, 3, 5] == 0
    assert cube[0, 2, 0, 0] == 0
    assert cube[0, 2, 1, 0] == DQFLAGS["JUMP_DET"]  # Jump was extended
    assert cube[0, 2, 2, 2] == DQFLAGS["SATURATED"]  # Saturation was extended
    assert cube[0, 2, 3, 6] == DQFLAGS["JUMP_DET"]


def test_flag_large_events_groupedsnowball():
    cube = np.zeros(shape=(1, 5, 7, 7), dtype=np.uint8)
    # cross of saturation surrounding by jump -> snowball
#    cube[0, 1, :, :] = DQFLAGS["JUMP_DET"]
#    cube[0, 2, :, :] = DQFLAGS["JUMP_DET"]
    cube[0, 2, 1:6, 1:6] = DQFLAGS["JUMP_DET"]
    cube[0, 1, 1:6, 1:6] = DQFLAGS["JUMP_DET"]

    cube[0, 2, 3, 3] = DQFLAGS["SATURATED"]
    cube[0, 2, 2, 3] = DQFLAGS["SATURATED"]
    cube[0, 2, 3, 4] = DQFLAGS["SATURATED"]
    cube[0, 2, 4, 3] = DQFLAGS["SATURATED"]
    cube[0, 2, 3, 2] = DQFLAGS["SATURATED"]
    outgdq, num_snowballs = flag_large_events(
        cube,
        DQFLAGS["JUMP_DET"],
        DQFLAGS["SATURATED"],
        min_sat_area=1,
        min_jump_area=6,
        expand_factor=1.9,
        edge_size=0,
        sat_required_snowball=True,
        min_sat_radius_extend=0.5,
        sat_expand=1.1,
    )

    assert outgdq[0, 2, 1, 0] == DQFLAGS["JUMP_DET"]  # Jump was extended
    assert outgdq[0, 2, 2, 2] == DQFLAGS["SATURATED"]  # Saturation was extended


def test_flag_large_events_withsnowball_noextension():
    cube = np.zeros(shape=(1, 5, 7, 7), dtype=np.uint8)
    # cross of saturation surrounding by jump -> snowball
    cube[0, 2, 3, 3] = DQFLAGS["SATURATED"]
    cube[0, 2, 2, 3] = DQFLAGS["SATURATED"]
    cube[0, 2, 3, 4] = DQFLAGS["SATURATED"]
    cube[0, 2, 4, 3] = DQFLAGS["SATURATED"]
    cube[0, 2, 3, 2] = DQFLAGS["SATURATED"]
    cube[0, 2, 1, 1:6] = DQFLAGS["JUMP_DET"]
    cube[0, 2, 5, 1:6] = DQFLAGS["JUMP_DET"]
    cube[0, 2, 1:6, 1] = DQFLAGS["JUMP_DET"]
    cube[0, 2, 1:6, 5] = DQFLAGS["JUMP_DET"]
    cube, num_snowballs = flag_large_events(
        cube,
        DQFLAGS["JUMP_DET"],
        DQFLAGS["SATURATED"],
        min_sat_area=1,
        min_jump_area=6,
        expand_factor=1.9,
        edge_size=0,
        sat_required_snowball=True,
        min_sat_radius_extend=0.5,
        sat_expand=1.1,
        max_extended_radius=1,
    )
    assert cube[0, 1, 2, 2] == 0
    assert cube[0, 1, 3, 5] == 0
    assert cube[0, 2, 0, 0] == 0
    assert cube[0, 2, 1, 0] == 0  # Jump was NOT extended due to max_extended_radius=1
    assert cube[0, 2, 2, 2] == 0  # Saturation was NOT extended due to max_extended_radius=1


def test_find_faint_extended(tmp_path):
    nint, ngrps, ncols, nrows = 1, 66, 25, 25
    data = np.zeros(shape=(nint, ngrps, nrows, ncols), dtype=np.float32)
    gdq = np.zeros_like(data, dtype=np.uint32)
    pdq = np.zeros(shape=(nrows, ncols), dtype=np.uint32)
    pdq[0, 0] = 1
    pdq[1, 1] = 2147483648
    gain = 4
    readnoise = np.ones(shape=(nrows, ncols), dtype=np.float32) * 6.0 * gain
    rng = np.random.default_rng(12345)
    data[0, 1:, 14:20, 15:20] = 6 * gain * 6.0 * np.sqrt(2)
    data = data + rng.normal(size=(nint, ngrps, nrows, ncols)) * readnoise
    fits.writeto(tmp_path / "data.fits", data, overwrite=True)
    gdq, num_showers = find_faint_extended(
        data,
        gdq,
        pdq,
        readnoise * np.sqrt(2),
        1,
        100,
        DQFLAGS,
        snr_threshold=1.2,
        min_shower_area=10,
        inner=1,
        outer=2.6,
        sat_flag=2,
        jump_flag=4,
        ellipse_expand=1.,
        num_grps_masked=1,
        max_shower_amplitude=10
    )
    #  Check that all the expected samples in group 2 are flagged as jump and
    #  that they are not flagged outside
    fits.writeto(tmp_path / "gdq.fits", gdq, overwrite=True)
#    assert num_showers == 1
    assert np.all(gdq[0, 1, 22, 14:23] == 0)
    assert gdq[0, 1, 16, 18] == DQFLAGS['JUMP_DET']
    assert np.all(gdq[0, 1, 11:22, 16:19] == DQFLAGS["JUMP_DET"])
    assert np.all(gdq[0, 1, 22, 16:19] == 0)
    assert np.all(gdq[0, 1, 10, 16:19] == 0)
    #  Check that the same area is flagged in the first group after the event
    assert np.all(gdq[0, 2, 22, 14:23] == 0)
    assert gdq[0, 2, 16, 18] == DQFLAGS['JUMP_DET']
    assert np.all(gdq[0, 2, 11:22, 16:19] == DQFLAGS["JUMP_DET"])
    assert np.all(gdq[0, 2, 22, 16:19] == 0)
    assert np.all(gdq[0, 2, 10, 16:19] == 0)

    assert np.all(gdq[0, 3:, :, :]) == 0

    #  Check that the flags are not applied in the 3rd group after the event
    assert np.all(gdq[0, 4, 12:22, 14:23]) == 0


# No shower is found because the event is identical in all ints
def test_find_faint_extended_sigclip():
    nint, ngrps, ncols, nrows = 101, 6, 30, 30
    data = np.zeros(shape=(nint, ngrps, nrows, ncols), dtype=np.float32)
    gdq = np.zeros_like(data, dtype=np.uint8)
    pdq = np.zeros(shape=(nrows, ncols), dtype=np.uint8)
    gain = 4
    readnoise = np.ones(shape=(nrows, ncols), dtype=np.float32) * 6.0 * gain
    rng = np.random.default_rng(12345)
    data[0, 1:, 14:20, 15:20] = 6 * gain * 1.7
    data = data + rng.normal(size=(nint, ngrps, nrows, ncols)) * readnoise
    min_shower_area=20
    gdq, num_showers = find_faint_extended(
        data,
        gdq,
        pdq,
        readnoise,
        1,
        100,
        DQFLAGS,
        snr_threshold=1.3,
        min_shower_area=min_shower_area,
        inner=1,
        outer=2,
        sat_flag=2,
        jump_flag=4,
        ellipse_expand=1.1,
        num_grps_masked=3,
    )
    #  Check that all the expected samples in group 2 are flagged as jump and
    #  that they are not flagged outside
    assert num_showers == 0
    assert np.all(gdq[0, 1, 22, 14:23] == 0)
    assert np.all(gdq[0, 1, 21, 16:20] == 0)
    assert np.all(gdq[0, 1, 20, 15:22] == 0)
    assert np.all(gdq[0, 1, 19, 15:23] == 0)
    assert np.all(gdq[0, 1, 18, 14:23] == 0)
    assert np.all(gdq[0, 1, 17, 14:23] == 0)
    assert np.all(gdq[0, 1, 16, 14:23] == 0)
    assert np.all(gdq[0, 1, 15, 14:22] == 0)
    assert np.all(gdq[0, 1, 14, 16:22] == 0)
    assert np.all(gdq[0, 1, 13, 17:21] == 0)
    assert np.all(gdq[0, 1, 12, 14:23] == 0)
    assert np.all(gdq[0, 1, 12:23, 24] == 0)
    assert np.all(gdq[0, 1, 12:23, 13] == 0)

    #  Check that the flags are not applied in the 3rd group after the event
    assert np.all(gdq[0, 4, 12:22, 14:23]) == 0

# No shower is found because the event is identical in all ints
def test_find_faint_extended_sigclip():
    nint, ngrps, ncols, nrows = 101, 6, 30, 30
    data = np.zeros(shape=(nint, ngrps, nrows, ncols), dtype=np.float32)
    gdq = np.zeros_like(data, dtype=np.uint8)
    pdq = np.zeros(shape=(nrows, ncols), dtype=np.int32)
    gain = 4
    readnoise = np.ones(shape=(nrows, ncols), dtype=np.float32) * 6.0 * gain
    rng = np.random.default_rng(12345)
    data[0, 1:, 14:20, 15:20] = 6 * gain * 1.7
    data = data + rng.normal(size=(nint, ngrps, nrows, ncols)) * readnoise
    gdq, num_showers = find_faint_extended(data, gdq, pdq, readnoise, 1, 100,
                                           snr_threshold=1.3,
                                           min_shower_area=20, inner=1,
                                           outer=2, sat_flag=2, jump_flag=4,
                                           ellipse_expand=1.1, num_grps_masked=3,
                                           dqflags=DQFLAGS)
    #  Check that all the expected samples in group 2 are flagged as jump and
    #  that they are not flagged outside
    assert (np.all(gdq[0, 1, 22, 14:23] == 0))
    assert (np.all(gdq[0, 1, 21, 16:20] == 0))
    assert (np.all(gdq[0, 1, 20, 15:22] == 0))
    assert (np.all(gdq[0, 1, 19, 15:23] == 0))
    assert (np.all(gdq[0, 1, 18, 14:23] == 0))
    assert (np.all(gdq[0, 1, 17, 14:23] == 0))
    assert (np.all(gdq[0, 1, 16, 14:23] == 0))
    assert (np.all(gdq[0, 1, 15, 14:22] == 0))
    assert (np.all(gdq[0, 1, 14, 16:22] == 0))
    assert (np.all(gdq[0, 1, 13, 17:21] == 0))
    assert (np.all(gdq[0, 1, 12, 14:23] == 0))
    assert (np.all(gdq[0, 1, 12:23, 24] == 0))
    assert (np.all(gdq[0, 1, 12:23, 13] == 0))

    #  Check that the flags are not applied in the 3rd group after the event
    assert (np.all(gdq[0, 4, 12:22, 14:23]) == 0)


def test_inside_ellipse5():
    ellipse = ((0, 0), (1, 2), -10)
    point = (1, 0.6)
    result = point_inside_ellipse(point, ellipse)
    assert result


def test_inside_ellipse4():
    ellipse = ((0, 0), (1, 2), 0)
    point = (1, 0.5)
    result = point_inside_ellipse(point, ellipse)
    assert result

def test_inside_ellipse6():
    ellipse = ((0, 0), (1, 2), 0)
    point = (3, 0.5)
    result = point_inside_ellipse(point, ellipse)
    assert not result

def test_inside_ellipes5():
    point = (1110.5, 870.5)
    ellipse = ((1111.0001220703125, 870.5000610351562), (10.60660171508789, 10.60660171508789), 45.0)
    result = point_inside_ellipse(point, ellipse)
    assert result

def test_calc_num_slices():
    n_rows = 20
    max_available_cores = 10
    assert calc_num_slices(n_rows, "none", max_available_cores) == 1
    assert calc_num_slices(n_rows, "half", max_available_cores) == 5
    assert calc_num_slices(n_rows, "3", max_available_cores) == 3
    assert calc_num_slices(n_rows, "7", max_available_cores) == 7
    assert calc_num_slices(n_rows, "21", max_available_cores) == 10
    assert calc_num_slices(n_rows, "quarter", max_available_cores) == 2
    assert calc_num_slices(n_rows, "7.5", max_available_cores) == 1
    assert calc_num_slices(n_rows, "one", max_available_cores) == 1
    assert calc_num_slices(n_rows, "-5", max_available_cores) == 1
    assert calc_num_slices(n_rows, "all", max_available_cores) == 10
    assert calc_num_slices(n_rows, "3/4", max_available_cores) == 1
    n_rows = 9
    assert calc_num_slices(n_rows, "21", max_available_cores) == 9


def test_find_last_grp():
    assert (find_last_grp(grp=5, ngrps=7, num_grps_masked=0) == 6)
    assert (find_last_grp(grp=5, ngrps=7, num_grps_masked=2) == 7)
    assert (find_last_grp(grp=5, ngrps=7, num_grps_masked=3) == 7)
    assert (find_last_grp(grp=5, ngrps=6, num_grps_masked=1) == 6)
    assert (find_last_grp(grp=5, ngrps=6, num_grps_masked=0) == 6)
    assert (find_last_grp(grp=5, ngrps=6, num_grps_masked=2) == 6)
    assert (find_last_grp(grp=5, ngrps=8, num_grps_masked=0) == 6)
    assert (find_last_grp(grp=5, ngrps=8, num_grps_masked=1) == 7)
    assert (find_last_grp(grp=5, ngrps=8, num_grps_masked=2) == 8)



#
#  Module for dark subtracting science data sets
#

import copy
import logging

import numpy as np

from . import dark_class

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


def do_correction(input_model, dark_model, dark_output=None):
    """
    Convert models to classes to remove internal dependence on data models.
    After the creation of internal classes, the dark current subtraction is
    done.

    Parameters
    ----------
    input_model : data model
        Input model assumed to be like a JWST RampModel.

    dark_model : data model
        Input model assumed to be like a JWST DarkModel.

    dark_output : str
        A path to denote where to save the dark model, if desired.

    Returns
    -------
    output_data : ScienceData
        dark-subtracted science data

    averaged_dark : DarkData
        New dark object with averaged frames
    """
    science_data = dark_class.ScienceData(input_model)
    dark_data = dark_class.DarkData(dark_model=dark_model)

    return do_correction_data(science_data, dark_data, dark_output)


def do_correction_data(science_data, dark_data, dark_output=None):
    """
    Short Summary
    -------------
    Execute all tasks for Dark Current Subtraction

    Parameters
    ----------
    science_data : ScienceData
        science data to be corrected

    dark_data : dark model object
        dark data

    dark_output : string
        file name in which to optionally save averaged dark data

    Returns
    -------
    output_data : ScienceData
        dark-subtracted science data

    averaged_dark : DarkData
        New dark object with averaged frames
    """
    # Save some data params for easy use later
    sci_nints = science_data.data.shape[0]
    sci_ngroups = science_data.data.shape[1]
    sci_nframes = science_data.exp_nframes
    sci_groupgap = science_data.exp_groupgap

    if len(dark_data.data.shape) == 4:
        drk_nints = dark_data.data.shape[0]
        drk_ngroups = dark_data.data.shape[1]
    else:
        drk_nints = 1
        drk_ngroups = dark_data.data.shape[0]

    drk_nframes = dark_data.exp_nframes
    drk_groupgap = dark_data.exp_groupgap

    log.info(
        "Science data nints=%d, ngroups=%d, nframes=%d, groupgap=%d",
        sci_nints,
        sci_ngroups,
        sci_nframes,
        sci_groupgap,
    )
    log.info(
        "Dark data nints=%d, ngroups=%d, nframes=%d, groupgap=%d",
        drk_nints,
        drk_ngroups,
        drk_nframes,
        drk_groupgap,
    )

    # Check that the number of groups in the science data does not exceed
    # the number of groups in the dark current array.
    sci_total_frames = sci_ngroups * sci_nframes + (sci_ngroups - 1) * sci_groupgap
    drk_total_frames = drk_ngroups * drk_nframes + (drk_ngroups - 1) * drk_groupgap
    if sci_total_frames > drk_total_frames:
        log.warning("Not enough data in dark reference file to match to science data.")
        log.warning("Input will be returned without subtracting dark current.")
        science_data.cal_step = "SKIPPED"
        out_data = copy.deepcopy(science_data)

        return out_data, None

    # Check that the value of nframes and groupgap in the dark
    # are not greater than those of the science data
    if drk_nframes > sci_nframes or drk_groupgap > sci_groupgap:
        log.warning(
            "The value of nframes or groupgap in the dark data is "
            "greater than that of the science data."
            "Input will be returned without subtracting dark current."
        )
        science_data.cal_step = "SKIPPED"
        out_data = copy.deepcopy(science_data)
        return out_data, None

    # Replace NaN's in the dark with zeros
    dark_data.data[np.isnan(dark_data.data)] = 0.0

    # Check whether the dark and science data have matching
    # nframes and groupgap settings.
    averaged_dark = None
    if sci_nframes == drk_nframes and sci_groupgap == drk_groupgap:
        # They match, so we can subtract the dark ref file data directly
        output_data = subtract_dark(science_data, dark_data)

        # If the user requested to have the dark file saved,
        # save the reference model as this file. This will
        # ensure consistency from the user's standpoint
        if dark_output is not None:
            averaged_dark = dark_data
            averaged_dark.save = True
            averaged_dark.output_name = dark_output

    else:
        # Create a frame-averaged version of the dark data to match
        # the nframes and groupgap settings of the science data.
        # If the data are from JWST/MIRI, the darks are integration-dependent
        # and we average them with a separate routine.

        if len(dark_data.data.shape) == 4:  # only MIRI uses 4-D darks
            averaged_dark = average_dark_frames_4d(
                dark_data, sci_nints, sci_ngroups, sci_nframes, sci_groupgap
            )
        else:
            averaged_dark = average_dark_frames_3d(dark_data, sci_ngroups, sci_nframes, sci_groupgap)

        # Save the frame-averaged dark data that was just created,
        # if requested by the user
        if dark_output is not None:
            averaged_dark.save = True
            averaged_dark.output_name = dark_output

        # Subtract the frame-averaged dark data from the science data
        output_data = subtract_dark(science_data, averaged_dark)

    output_data.cal_step = "COMPLETE"

    return output_data, averaged_dark


def average_dark_frames_3d(dark_data, ngroups, nframes, groupgap):
    """
    Averages the individual frames of data in a dark reference
    file to match the group structure of a science data set.
    This routine is not used for JWST/MIRI (see average_dark_frames_4d).

    Parameters
    ----------
    dark_data : DarkData
        the input dark data

    ngroups : int
        number of groups in the science data set

    nframes : int
        number of frames per group in the science data set

    groupgap : int
        number of frames skipped between groups in the science data set

    Returns
    -------
    avg_dark : DarkData
        New dark object with averaged frames
    """
    # Create a model for the averaged dark data
    dny = dark_data.data.shape[1]
    dnx = dark_data.data.shape[2]
    dims = (ngroups, dny, dnx)
    avg_dark = dark_class.DarkData(dims=dims)

    # Do a direct copy of the 2-d DQ array into the new dark
    avg_dark.groupdq = dark_data.groupdq

    # Loop over the groups of the input science data, copying or
    # averaging the dark frames to match the group structure
    start = 0

    for group in range(ngroups):
        end = start + nframes

        # If there's only 1 frame per group, just copy the dark frames
        if nframes == 1:
            log.debug("copy dark frame %d", start)
            avg_dark.data[group] = dark_data.data[start]
            avg_dark.err[group] = dark_data.err[start]

        # Otherwise average nframes into a new group: take the mean of
        # the SCI arrays and the quadratic sum of the ERR arrays.
        else:
            log.debug("average dark frames %d to %d", start + 1, end)
            avg_dark.data[group] = dark_data.data[start:end].mean(axis=0)
            avg_dark.err[group] = np.sqrt(np.add.reduce(dark_data.err[start:end] ** 2, axis=0)) / (
                end - start
            )

        # Skip over unused frames
        start = end + groupgap

    # Reset some metadata values for the averaged dark
    avg_dark.exp_nframes = nframes
    avg_dark.exp_ngroups = ngroups
    avg_dark.exp_groupgap = groupgap

    return avg_dark


def average_dark_frames_4d(dark_data, nints, ngroups, nframes, groupgap):
    """
    Averages the individual frames of data in a dark reference
    file to match the group structure of a science data set.
    JWST/MIRI needs a separate routine because the darks are
    integration-dependent and hence 4D in shape, instead of 3D.
    An average dark is created for each integration.

    Parameters
    ----------
    dark_data : DarkData
        the input dark data

    nints : int
        number of integrations in the science data

    ngroups : int
        number of groups in the science data set

    nframes : int
        number of frames per group in the science data set

    groupgap : int
        number of frames skipped between groups in the science data set

    Returns
    -------
    avg_dark : dark data model
        New dark object with averaged frames
    """
    # Create a model for the averaged dark data
    dint = dark_data.data.shape[0]
    dny = dark_data.data.shape[2]
    dnx = dark_data.data.shape[3]
    dims = (dint, ngroups, dny, dnx)
    avg_dark = dark_class.DarkData(dims)

    # Do a direct copy of the 2-d DQ array into the new dark
    avg_dark.groupdq = dark_data.groupdq

    # check if the number of integrations in dark reference file
    # is less than science data.  If so then we only need to find the
    # average for dint integrations, otherwise we find the average for
    # nints.  (if science data only has 1 integration then there is no
    # need to average the second integration of the dark refefence file)
    num_ints = dint
    if dint > nints:
        num_ints = nints

    # Loop over valid integrations in dark reference file
    # Loop over the groups of the input science data, copying or
    # averaging the dark frames to match the group structure

    for it in range(num_ints):
        start = 0
        for group in range(ngroups):
            end = start + nframes

            # If there's only 1 frame per group, just copy the dark frames
            if nframes == 1:
                log.debug("copy dark frame %d", start)
                avg_dark.data[it, group] = dark_data.data[it, start]
                avg_dark.err[it, group] = dark_data.err[it, start]

            # Otherwise average nframes into a new group: take the mean of
            # the SCI arrays and the quadratic sum of the ERR arrays.
            else:
                log.debug("average dark frames %d to %d", start + 1, end)
                avg_dark.data[it, group] = dark_data.data[it, start:end].mean(axis=0)
                avg_dark.err[it, group] = np.sqrt(
                    np.add.reduce(dark_data.err[it, start:end] ** 2, axis=0)
                ) / (end - start)

            # Skip over unused frames
            start = end + groupgap

    # Reset some metadata values for the averaged dark
    avg_dark.exp_nframes = nframes
    avg_dark.exp_ngroups = ngroups
    avg_dark.exp_groupgap = groupgap

    return avg_dark


def subtract_dark(science_data, dark_data):
    """
    Subtracts dark current data from science arrays, combines
    error arrays in quadrature, and updates data quality array based on
    DQ flags in the dark arrays.

    Parameters
    ----------
    science_data : ScienceData
        the input science data

    dark_data : DarkData
        the dark current data

    Returns
    -------
    output : data model object
        dark-subtracted science data
    """
    # The integration start number is only needed for JWST/MIRI data.
    # It defaults to 1 if the keyword is not in the science data.
    int_start = 1 if science_data.exp_intstart is None else science_data.exp_intstart

    # Determine the number of integrations contained in the dark reference file
    dark_nints = dark_data.data.shape[0] if len(dark_data.data.shape) == 4 else 1

    log.debug(
        "subtract_dark: nints=%d, ngroups=%d, size=%d,%d",
        science_data.data.shape[0],
        science_data.data.shape[1],
        science_data.data.shape[2],
        science_data.data.shape[3],
    )

    # Create output as a copy of the input science data model
    output = copy.deepcopy(science_data)

    if len(dark_data.data.shape) == 4:
        # MIRI dark reference file has a DQ plane for each integration,
        # so we collapse the dark DQ planes into a single 2-D array
        darkdq = dark_data.groupdq[0, 0, :, :].copy()
        for i in range(1, dark_nints):
            darkdq = np.bitwise_or(darkdq, dark_data.groupdq[i, 0, :, :])
    else:
        # All other instruments have a single 2D dark DQ array
        darkdq = dark_data.groupdq

    # Propagate the dark DQ data into the science DQ data
    output.pixeldq = np.bitwise_or(science_data.pixeldq, darkdq)

    # Loop over all integrations in input science data
    for i in range(science_data.data.shape[0]):
        if len(dark_data.data.shape) == 4:  # MIRI data
            # Apply the first dark_nints-1 integrations from the dark ref file
            # to the first few science integrations. There's an additional
            # check of the starting integration number in case the science
            # data are segmented.
            #
            # else
            #
            # For science integrations beyond the number of
            # dark integrations, use the last dark integration
            dark_sci = dark_data.data[i] if i < dark_nints and int_start == 1 else dark_data.data[-1]
        else:
            # Use single-integration dark data
            dark_sci = dark_data.data

        # Loop over all groups in this integration
        for j in range(science_data.data.shape[1]):
            # subtract the dark from the science data
            output.data[i, j] -= dark_sci[j]

    return output



"""Test astrometric utility functions for alignment"""

import copy
from copy import deepcopy
from pathlib import Path

import asdf
import contextlib
import numpy as np
import pytest
from astropy.modeling.models import Shift
from astropy.table import Table
from astropy.time import Time

from stcal.tweakreg import astrometric_utils as amutils
from stcal.tweakreg.tweakreg import (
    TweakregError,
    _is_wcs_correction_small,
    _parse_refcat,
    _wcs_to_skycoord,
    absolute_align,
    construct_wcs_corrector,
    relative_align,
)
from stcal.tweakreg.utils import _wcsinfo_from_wcs_transform
import requests

# Define input GWCS specification to be used for these tests
WCS_NAME = "mosaic_long_i2d_gwcs.asdf"  # Derived using B7.5 Level 3 product
EXPECTED_NUM_SOURCES = 2469

# more recent WCS with a defined input frame is necessary for some tests
WCS_NAME_2 = "nrcb1-wcs.asdf"

TEST_CATALOG = "GAIADR3"
CATALOG_FNAME = "ref_cat.ecsv"
DATADIR = "data"

# something
BKG_LEVEL = 0.001
N_EXAMPLE_SOURCES = 21


class MockConnectionError:
    def __init__(self, *args, **kwargs):
        raise requests.exceptions.ConnectionError


@pytest.fixture(scope="module")
def wcsobj():
    path = Path(__file__).parent / DATADIR / WCS_NAME
    with asdf.open(path, lazy_load=False) as asdf_file:
        return asdf_file.tree["wcs"]


@pytest.fixture(scope="module")
def wcsobj2():
    path = Path(__file__).parent / DATADIR / WCS_NAME_2
    with asdf.open(path, lazy_load=False) as asdf_file:
        return asdf_file.tree["wcs"]


def test_radius(wcsobj):
    # compute radius
    radius, _ = amutils.compute_radius(wcsobj)

    # check results
    EXPECTED_RADIUS = 0.02564497890604383
    np.testing.assert_allclose(radius, EXPECTED_RADIUS, rtol=1e-6)


def test_get_catalog(wcsobj):
    # Get radius and fiducial
    radius, fiducial = amutils.compute_radius(wcsobj)

    # Get the catalog
    cat = amutils.get_catalog(fiducial[0], fiducial[1], search_radius=radius,
                              catalog=TEST_CATALOG)

    assert len(cat) == EXPECTED_NUM_SOURCES


def test_create_catalog(wcsobj):
    # Create catalog
    gcat = amutils.create_astrometric_catalog(
        wcsobj, "2016.0",
        catalog=TEST_CATALOG,
        output=None,
    )
    # check that we got expected number of sources
    assert len(gcat) == EXPECTED_NUM_SOURCES


def test_create_catalog_graceful_failure(wcsobj):
    """
    Ensure catalog retuns zero sources instead of failing outright
    when the bounding box is too small to find any sources
    """
    wcsobj.bounding_box = ((0, 0.5), (0, 0.5))

    # Create catalog
    gcat = amutils.create_astrometric_catalog(
        wcsobj, "2016.0",
        catalog=TEST_CATALOG,
        output=None,
    )
    # check that we got expected number of sources
    assert len(gcat) == 0


def fake_correctors(offset):
    path = Path(__file__).parent / DATADIR / WCS_NAME
    with asdf.open(path) as af:
        wcs = af.tree["wcs"]

    # Make a copy and add an offset at the end of the transform
    twcs = deepcopy(wcs)
    step = twcs.pipeline[0]
    step.transform = step.transform | Shift(offset) & Shift(offset)
    twcs.bounding_box = wcs.bounding_box

    class FakeCorrector:
        def __init__(self, wcs, original_skycoord):
            self.wcs = wcs
            self._original_skycoord = original_skycoord

        @property
        def meta(self):
            return {"original_skycoord": self._original_skycoord}

    return [FakeCorrector(twcs, _wcs_to_skycoord(wcs))]


@pytest.mark.parametrize(("offset", "is_good"),
                         [(1 / 3600, True), (11 / 3600, False)])
def test_is_wcs_correction_small(offset, is_good):
    """
    Test that the _is_wcs_correction_small method returns True for a small
    wcs correction and False for a "large" wcs correction. The values in this
    test are selected based on the current step default parameters:
        - use2dhist
        - searchrad
        - tolerance
    Changes to the defaults for these parameters will likely require updating
    the values uses for parametrizing this test.
    """
    correctors = fake_correctors(offset)
    assert _is_wcs_correction_small(correctors) == is_good


def test_expected_fails_bad_separation():

    correctors = fake_correctors(0.0)
    separation = 1.0
    tolerance = 1.0
    with pytest.raises(TweakregError):
        relative_align(correctors,
                       separation=separation,
                       tolerance=tolerance)

    with pytest.raises(TweakregError):
        absolute_align(correctors, "GAIADR3",
                       None,
                       None,
                       None,
                       abs_separation=separation,
                       abs_tolerance=tolerance)


class AttrDict(dict):
    """Hack to be able to treat wcsinfo dict as an object so attributes
    can be accessed"""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__dict__ = self

    @property
    def instance(self):
        return self


class Metadata:

    def __init__(self, wcs, epoch, group_id=None):
        self.wcs = wcs
        self.observation = AttrDict({"date": epoch})
        wcsinfo = _wcsinfo_from_wcs_transform(wcs)
        wcsinfo["v3yangle"] = 0.0
        wcsinfo["vparity"] = 1
        self.wcsinfo = AttrDict(wcsinfo)
        self.group_id = group_id


class MinimalDataWithWCS:

    def __init__(self, wcs, epoch="2016-01-01T00:00:00.0", group_id=None):
        self.meta = Metadata(wcs, epoch, group_id=group_id)
        self.data = np.zeros((512, 512))


@pytest.fixture(scope="module")
def datamodel(wcsobj2, group_id=None):
    return MinimalDataWithWCS(wcsobj2, group_id=group_id)


def test_parse_refcat(datamodel, tmp_path):

    wcsobj = datamodel.meta.wcs
    correctors = fake_correctors(0.0)

    # Get radius and fiducial
    radius, fiducial = amutils.compute_radius(wcsobj)

    # Get the catalog
    cat = amutils.get_catalog(fiducial[0], fiducial[1], search_radius=radius,
                              catalog=TEST_CATALOG)

    # save refcat to file
    cat.write(tmp_path / CATALOG_FNAME, format="ascii.ecsv", overwrite=True)

    # parse refcat from file
    epoch = Time(datamodel.meta.observation.date).decimalyear
    refcat = _parse_refcat(tmp_path / CATALOG_FNAME,
                           correctors,
                           datamodel.meta.wcs,
                           datamodel.meta.wcsinfo,
                           epoch)
    assert isinstance(refcat, Table)

    # find refcat from web
    refcat = _parse_refcat(TEST_CATALOG, correctors, datamodel.meta.wcs, datamodel.meta.wcsinfo, epoch)
    assert isinstance(refcat, Table)


@pytest.fixture(scope="module")
def input_catalog(datamodel):
    """Get catalog from gaia, transform it to x,y in the image frame,
    use it as an input catalog"""
    # Get radius and fiducial
    w = datamodel.meta.wcs
    radius, fiducial = amutils.compute_radius(w)

    # Get the catalog
    cat = amutils.get_catalog(fiducial[0], fiducial[1], search_radius=radius,
                              catalog=TEST_CATALOG)

    x, y = w.world_to_pixel(cat["ra"].value, cat["dec"].value)
    return Table({"x": x, "y": y})


@pytest.fixture(scope="module")
def example_input(wcsobj2):

    m0 = MinimalDataWithWCS(wcsobj2)
    m0.data[:] = BKG_LEVEL
    n_sources = N_EXAMPLE_SOURCES
    rng = np.random.default_rng(26)
    xs = rng.choice(50, n_sources, replace=False) * 8 + 10
    ys = rng.choice(50, n_sources, replace=False) * 8 + 10
    for y, x in zip(ys, xs, strict=False):
        m0.data[y-1:y+2, x-1:x+2] = [
            [0.1, 0.6, 0.1],
            [0.6, 0.8, 0.6],
            [0.1, 0.6, 0.1],
        ]

    m1 = copy.deepcopy(m0)
    # give each a unique filename
    m0.meta.filename = "some_file_0.fits"
    m0.meta.group_id = "a"
    m1.meta.filename = "some_file_1.fits"
    m1.meta.group_id = "b"
    return [m0, m1]


@pytest.mark.parametrize("with_shift", [True, False])
def test_relative_align(example_input, input_catalog, with_shift):

    [m0, m1] = example_input
    cat1 = copy.deepcopy(input_catalog)
    if with_shift:
        m1.data[:-9] = m0.data[9:]
        m1.data[-9:] = BKG_LEVEL
        cat1["y"] -= 9

    correctors = [construct_wcs_corrector(dm.meta.wcs,
                                          dm.meta.wcsinfo.instance,
                                          cat,
                                          dm.meta.group_id) for (dm, cat) in \
                                          zip([m0, m1], [input_catalog, cat1], strict=True)]
    result = relative_align(correctors, minobj=5)

    # ensure wcses differ by a small amount due to the shift above
    # by projecting one point through each wcs and comparing the difference
    abs_delta = abs(result[1].wcs(0, 0)[0] - result[0].wcs(0, 0)[0])
    if with_shift:
        assert abs_delta > 1E-5
    else:
        assert abs_delta < 1E-12


def test_absolute_align(example_input, input_catalog):

    correctors = [construct_wcs_corrector(dm.meta.wcs,
                                          dm.meta.wcsinfo.instance,
                                          input_catalog,
                                          dm.meta.group_id) for dm in example_input]

    ref_model = example_input[0]
    result = absolute_align(correctors,
                            TEST_CATALOG,
                            ref_wcs=ref_model.meta.wcs,
                            ref_wcsinfo=ref_model.meta.wcsinfo,
                            epoch=Time(ref_model.meta.observation.date).decimalyear,
                            abs_minobj=5)
    for res in result:
        assert res.meta["group_id"] == 987654

    abs_delta = abs(result[1].wcs(0, 0)[0] - result[0].wcs(0, 0)[0])
    assert abs_delta < 1E-12

def test_get_catalog_timeout():
    """Test that get_catalog can raise an exception on timeout."""

    with pytest.raises(Exception) as exec_info:
        for dt in np.arange(1, 0, -0.01):
            with contextlib.suppress(requests.exceptions.ConnectionError):
                amutils.get_catalog(10, 10, search_radius=0.1, catalog="GAIADR3", timeout=dt)
    assert exec_info.type == requests.exceptions.Timeout


def test_get_catalog_raises_connection_error(monkeypatch):
    """Test that get_catalog can raise an exception on connection error."""

    monkeypatch.setattr("requests.get", MockConnectionError)

    with pytest.raises(Exception) as exec_info:
        amutils.get_catalog(10, 10, search_radius=0.1, catalog="GAIADR3")

    assert exec_info.type == requests.exceptions.ConnectionError



"""

Unit tests for linearity correction

"""

import numpy as np

from stcal.linearity.linearity import linearity_correction

DQFLAGS = {"GOOD": 0, "DO_NOT_USE": 1, "SATURATED": 2, "DEAD": 1024, "HOT": 2048, "NO_LIN_CORR": 1048576}

DELIM = "-" * 80


def test_coeff_dq():
    """
    Test linearity algorithm with random data ramp (does
    algorithm match expected algorithm) also test a variety
    of dq flags and expected output
    """

    # size of integration
    nints = 1
    ngroups = 160
    xsize = 103
    ysize = 102

    # Create data array and group/pixel dq arrays
    data = np.ones((nints, ngroups, ysize, xsize)).astype(np.float32)
    pdq = np.zeros((ysize, xsize)).astype(np.uint32)
    gdq = np.zeros((nints, ngroups, ysize, xsize)).astype(np.uint32)

    # Create reference file data and dq arrays
    numcoeffs = 5
    lin_coeffs = np.zeros((numcoeffs, ysize, xsize))

    # Set coefficient values in reference file to check the algorithm
    # Equation is DNcorr = L0 + L1*DN(i) + L2*DN(i)^2 + L3*DN(i)^3 + L4*DN(i)^4
    # DN(i) = signal in pixel, Ln = coefficient from ref file
    # L0 = 0 for all pixels for CDP6
    L0 = 0.0e00
    L1 = 0.85
    L2 = 4.62e-6
    L3 = -6.16e-11
    L4 = 7.23e-16

    coeffs = np.asarray([L0, L1, L2, L3, L4], dtype="float")

    # pixels we are testing using above coefficients
    lin_coeffs[:, 30, 50] = coeffs
    lin_coeffs[:, 35, 36] = coeffs
    lin_coeffs[:, 35, 35] = coeffs

    lin_dq = np.zeros((ysize, xsize), dtype=np.uint32)

    # check behavior with NaN coefficients: should not alter pixel values
    coeffs2 = np.asarray([L0, np.nan, L2, L3, L4], dtype="float")

    lin_coeffs[:, 20, 50] = coeffs2
    data[0, 50, 20, 50] = 500.0

    # test case where all coefficients are zero
    lin_coeffs[:, 25, 25] = 0.0
    data[0, 50, 25, 25] = 600.0

    tgroup = 2.775

    # set pixel values (DN) for specific pixels up the ramp
    data[0, :, 30, 50] = np.arange(ngroups) * 100 * tgroup

    scival = 40000.0
    data[0, 45, 30, 50] = scival  # to check linearity multiplication is done correctly
    data[0, 30, 35, 36] = 35  # pixel to check that dq=2 meant no correction was applied

    # check if dq flags in pixeldq are correctly populated in output
    pdq[50, 40] = DQFLAGS["DO_NOT_USE"]
    pdq[50, 41] = DQFLAGS["SATURATED"]
    pdq[50, 42] = DQFLAGS["DEAD"]
    pdq[50, 43] = DQFLAGS["HOT"]

    # set dq flags in DQ of reference file
    lin_dq[35, 35] = DQFLAGS["DO_NOT_USE"]
    lin_dq[35, 36] = DQFLAGS["NO_LIN_CORR"]
    lin_dq[30, 50] = DQFLAGS["GOOD"]

    np.bitwise_or(pdq, lin_dq)

    # run linearity correction
    output_data, output_pdq, _ = linearity_correction(data, gdq, pdq, lin_coeffs, lin_dq, DQFLAGS)

    # check that multiplication of polynomial was done correctly for specified pixel
    outval = L0 + (L1 * scival) + (L2 * scival**2) + (L3 * scival**3) + (L4 * scival**4)
    assert np.isclose(output_data[0, 45, 30, 50], outval, rtol=0.00001)

    # check that dq value was handled correctly

    assert output_pdq[35, 35] == DQFLAGS["DO_NOT_USE"]
    assert output_pdq[35, 36] == DQFLAGS["NO_LIN_CORR"]
    # NO_LIN_CORR, sci value should not change
    assert output_data[0, 30, 35, 36] == 35
    # NaN coefficient should not change data value
    assert output_data[0, 50, 20, 50] == 500.0
    # dq for pixel with all zero lin coeffs should be NO_LIN_CORR
    assert output_pdq[25, 25] == DQFLAGS["NO_LIN_CORR"]


def create_science_data(dims, ncoeffs):
    """
    Create science data arrays with specific dimensions.

    dims : tuple
        The dimensions of the science data (nints, ngroups, nrows, ncols).

    ncoeffs : int
        The number of coefficients for the linear correction.
    """
    nints, ngroups, nrows, ncols = dims
    image_shape = (nrows, ncols)
    coeffs_shape = (ncoeffs, nrows, ncols)

    data = np.zeros(dims, dtype=float)
    gdq = np.zeros(dims, dtype=np.uint8)
    pdq = np.zeros(image_shape, dtype=np.uint8)
    zframe = np.zeros((nints, nrows, ncols), dtype=float)

    lin_coeffs = np.zeros(coeffs_shape, dtype=float)
    lin_dq = np.zeros(image_shape, dtype=np.uint32)

    return data, gdq, pdq, lin_coeffs, lin_dq, zframe


def test_zero_frame():
    """
    Check to make sure the ZEROFRAME properly gets corrected.
    """

    nints, ngroups, nrows, ncols = 1, 5, 1, 2
    ncoeffs = 5
    dims = nints, ngroups, nrows, ncols
    ncoeffs = 5

    data, gdq, pdq, lin_coeffs, lin_dq, zframe = create_science_data(dims, ncoeffs)

    base = 31.459
    data[0, :, 0, 0] = np.array([(k + 1) * base for k in range(ngroups)], dtype=float)
    zframe[0, 0, :] = np.array([data[0, 0, 0, 0] * 0.666666, 0.0])

    lin_base = 2.718 / (base * 10.0)
    coeffs = np.array([lin_base ** (k) for k in range(ncoeffs)], dtype=float)
    lin_coeffs[:, 0, 0] = coeffs

    output_data, output_pdq, new_zframe = linearity_correction(
        data, gdq, pdq, lin_coeffs, lin_dq, DQFLAGS, zframe
    )

    zcheck = np.zeros((nints, nrows, ncols), dtype=float)
    zcheck[0, 0, :] = np.array([1.22106063, 0.0])
    np.testing.assert_almost_equal(new_zframe, zcheck, decimal=5)



# cython: language_level=3str

"""
External interface module for the Casertano+22 ramp fitting algorithm with jump detection.
    This module is intended to contain everything needed by external code.

Enums
-----
Parameter :
    Enumerate the index for the output parameters array.

Variance :
    Enumerate the index for the output variances array.

Classes
-------
RampFitOutputs : NamedTuple
    Simple tuple wrapper for outputs from the ramp fitting algorithm
        This clarifies the meaning of the outputs via naming them something
        descriptive.

(Public) Functions
------------------
fit_ramps : function
    Fit ramps using the Castenario+22 algorithm to a set of pixels accounting
    for jumps (if use_jump is True) and bad pixels (via the dq array). This
    is the primary externally callable function.
"""
from __future__ import annotations

import numpy as np

cimport numpy as cnp
from cython cimport boundscheck, wraparound
from libcpp cimport bool
from libcpp.list cimport list as cpp_list

from stcal.ramp_fitting.ols_cas22._jump cimport (
    JumpFits,
    Thresh,
    fill_fixed_values,
    fit_jumps,
    n_fixed_offsets,
    n_pixel_offsets,
)
from stcal.ramp_fitting.ols_cas22._ramp cimport ReadPattern, from_read_pattern

from typing import NamedTuple

# Initialize numpy for cython use in this module
cnp.import_array()


cpdef enum Parameter:
    intercept
    slope
    n_param


cpdef enum Variance:
    read_var
    poisson_var
    total_var
    n_var


class RampFitOutputs(NamedTuple):
    """
    Simple tuple wrapper for outputs from the ramp fitting algorithm
        This clarifies the meaning of the outputs via naming them something
        descriptive.

    Attributes
    ----------
        parameters: np.ndarray[n_pixel, 2]
            the slope and intercept for each pixel's ramp fit. see Parameter enum
            for indexing indicating slope/intercept in the second dimension.
        variances: np.ndarray[n_pixel, 3]
            the read, poisson, and total variances for each pixel's ramp fit.
            see Variance enum for indexing indicating read/poisson/total in the
            second dimension.
        dq: np.ndarray[n_resultants, n_pixel]
            the dq array, with additional flags set for jumps detected by the
            jump detection algorithm.
        fits: list of RampFits
            the raw ramp fit outputs, these are all structs which will get mapped to
            python dictionaries.
    """
    parameters: np.ndarray
    variances: np.ndarray
    dq: np.ndarray
    fits: list | None = None


@boundscheck(False)
@wraparound(False)
def fit_ramps(float[:, :] resultants,
              cnp.ndarray[int, ndim=2] dq,
              float[:] read_noise,
              float read_time,
              list[list[int]] read_pattern,
              bool use_jump=False,
              float intercept=5.5,
              float constant=1/3,
              bool include_diagnostic=False):
    """Fit ramps using the Casertano+22 algorithm.
        This implementation uses the Cas22 algorithm to fit ramps, where
        ramps are fit between bad resultants marked by dq flags for each pixel
        which are not equal to zero. If use_jump is True, it additionally uses
        jump detection to mark additional resultants for each pixel as bad if
        a jump is suspected in them.

    Parameters
    ----------
    resultants : float[n_resultants, n_pixel]
        the resultants in electrons (Note that this can be based as any sort of
        array, such as a numpy array. The memory view is just for efficiency in
        cython)
    dq : np.ndarry[n_resultants, n_pixel]
        the dq array.  dq != 0 implies bad pixel / CR. (Kept as a numpy array
        so that it can be passed out without copying into new numpy array, will
        be working on memory views of this array)
    read_noise : float[n_pixel]
        the read noise in electrons for each pixel (same note as the resultants)
    read_time : float
        Time to perform a readout. For Roman data, this is FRAME_TIME.
    read_pattern : list[list[int]]
        the read pattern for the image
    use_jump : bool
        If True, use the jump detection algorithm to identify CRs.
        If False, use the DQ array to identify CRs.
    intercept : float
        The intercept value for the threshold function. Default=5.5
    constant : float
        The constant value for the threshold function. Default=1/3.0
    include_diagnostic : bool
        If True, include the raw ramp fits in the output. Default=False

    Returns
    -------
    A RampFitOutputs tuple
    """
    cdef int n_pixels, n_resultants
    n_resultants = resultants.shape[0]
    n_pixels = resultants.shape[1]

    # Raise error if input data is inconsistent
    if n_resultants != len(read_pattern):
        raise RuntimeError(f'The read pattern length {len(read_pattern)} does not '
                           f'match number of resultants {n_resultants}')

    # Compute the main metadata from the read pattern and cast it to memory views
    cdef ReadPattern metadata = from_read_pattern(read_pattern, read_time, n_resultants)
    cdef float[:] t_bar = metadata.t_bar
    cdef float[:] tau = metadata.tau
    cdef int[:] n_reads = metadata.n_reads

    # Setup pre-compute arrays for jump detection
    cdef float[:, :] fixed
    cdef float[:, :] pixel
    if use_jump:
        # Initialize arrays for the jump detection pre-computed values
        fixed = np.empty((n_fixed_offsets, n_resultants - 1), dtype=np.float32)
        pixel = np.empty((n_pixel_offsets, n_resultants - 1), dtype=np.float32)

        # Pre-compute the values from the read pattern
        fixed = fill_fixed_values(fixed, t_bar, tau, n_reads, n_resultants)
    else:
        # "Initialize" the arrays when not using jump detection, they need to be
        #    initialized because they do get passed around, but they don't need
        #    to actually have any entries
        fixed = np.empty((0, 0), dtype=np.float32)
        pixel = np.empty((0, 0), dtype=np.float32)

    # Create a threshold struct
    cdef Thresh thresh = Thresh(intercept, constant)

    # Create variable to old the diagnostic data
    # Use list because this might grow very large which would require constant
    #    reallocation. We don't need random access, and this gets cast to a python
    #    list in the end.
    cdef cpp_list[JumpFits] ramp_fits

    # Initialize the output arrays. Note that the fit intercept is currently always
    #    zero, where as every variance is calculated and set. This means that the
    #    parameters need to be filled with zeros, where as the variances can just
    #    be allocated
    cdef float[:, :] parameters = np.zeros((n_pixels, Parameter.n_param), dtype=np.float32)
    cdef float[:, :] variances = np.empty((n_pixels, Variance.n_var), dtype=np.float32)

    # Cast the enum values into integers for indexing (otherwise compiler complains)
    #   These will be optimized out
    cdef int slope = Parameter.slope
    cdef int read_var = Variance.read_var
    cdef int poisson_var = Variance.poisson_var
    cdef int total_var = Variance.total_var

    # Pull memory view of dq for speed of access later
    #   changes to this array will backpropagate to the original numpy array
    cdef int[:, :] dq_ = dq

    # Run the jump fitting algorithm for each pixel
    cdef JumpFits fit
    cdef int index
    for index in range(n_pixels):
        # Fit all the ramps for the given pixel
        fit = fit_jumps(resultants[:, index],
                        dq_[:, index],
                        read_noise[index],
                        t_bar,
                        tau,
                        n_reads,
                        n_resultants,
                        fixed,
                        pixel,
                        thresh,
                        use_jump,
                        include_diagnostic)

        # Extract the output fit's parameters
        parameters[index, slope] = fit.average.slope

        # Extract the output fit's variances
        variances[index, read_var] = fit.average.read_var
        variances[index, poisson_var] = fit.average.poisson_var
        variances[index, total_var] = fit.average.read_var + fit.average.poisson_var

        # Store diagnostic data if requested
        if include_diagnostic:
            ramp_fits.push_back(fit)

    # Cast memory views into numpy arrays for ease of use in python.
    return RampFitOutputs(np.array(parameters, dtype=np.float32),
                          np.array(variances, dtype=np.float32),
                          dq,
                          ramp_fits if include_diagnostic else None)



import warnings

import gwcs
import pytest
import numpy as np
import scipy.signal
from astropy.modeling import models

from stcal.outlier_detection.utils import (
    _abs_deriv,
    compute_weight_threshold,
    flag_crs,
    flag_resampled_crs,
    gwcs_blot,
    calc_gwcs_pixmap,
    reproject,
    medfilt,
)
from stcal.testing_helpers import MemoryThreshold


@pytest.mark.parametrize("shape,diff", [
    ([5, 7], 100),
    ([17, 13], -200),
])
def test_abs_deriv_single_value(shape, diff):
    arr = np.zeros(shape)
    # put diff at the center
    np.put(arr, arr.size // 2, diff)
    # since abs_deriv with a single non-zero value is the same as a
    # convolution with a 3x3 cross kernel use it to test the result
    expected = scipy.signal.convolve2d(np.abs(arr), [[0, 1, 0], [1, 1, 1], [0, 1, 0]], mode='same')
    result = _abs_deriv(arr)
    np.testing.assert_allclose(result, expected)


@pytest.mark.parametrize("nrows,ncols", [(5, 5), (7, 11), (17, 13)])
def test_abs_deriv_range(nrows, ncols):
    arr = np.arange(nrows * ncols).reshape(nrows, ncols)
    result = _abs_deriv(arr)
    np.testing.assert_allclose(result, ncols)


def test_abs_deriv_nan():
    arr = np.arange(25, dtype='f4').reshape(5, 5)
    arr[2, 2] = np.nan
    expect_nan = np.zeros_like(arr, dtype=bool)
    expect_nan[2, 2] = True
    result = _abs_deriv(arr)
    assert np.isnan(result[expect_nan])
    assert np.all(np.isfinite(result[~expect_nan]))


@pytest.mark.parametrize("shape,mean,maskpt,expected", [
    ([5, 5], 11, 0.5, 5.5),
    ([5, 5], 11, 0.25, 2.75),
    ([3, 3, 3], 17, 0.5, 8.5),
])
def test_compute_weight_threshold(shape, mean, maskpt, expected):
    arr = np.ones(shape, dtype=np.float32) * mean
    result = compute_weight_threshold(arr, maskpt)
    np.testing.assert_allclose(result, expected)


def test_compute_weight_threshold_outlier():
    """
    Test that a large outlier doesn't bias the threshold
    """
    arr = np.ones([7, 7, 7], dtype=np.float32) * 42
    arr[3, 3] = 9000
    result = compute_weight_threshold(arr, 0.5)
    np.testing.assert_allclose(result, 21)


def test_compute_weight_threshold_zeros():
    """
    Test that zeros are ignored
    """
    arr = np.zeros([10, 10], dtype=np.float32)
    arr[:5, :5] = 42
    result = compute_weight_threshold(arr, 0.5)
    np.testing.assert_allclose(result, 21)


def test_compute_weight_threshold_memory():
    """Test that weight threshold function modifies
    the weight array in place"""
    arr = np.zeros([500, 500], dtype=np.float32)
    arr[:250, :250] = 42
    arr[10,10] = 0
    arr[-10,-10] = np.nan

    # buffer to account for memory overhead needs to be small enough
    # to ensure that the array was not copied
    fractional_memory_buffer = 0.9
    expected_mem = int(arr.nbytes*fractional_memory_buffer)
    with MemoryThreshold(str(expected_mem) + " B"):
        result = compute_weight_threshold(arr, 0.5)
    np.testing.assert_allclose(result, 21)


def test_flag_crs():
    sci = np.zeros((10, 10), dtype=np.float32)
    err = np.ones_like(sci)
    blot = np.zeros_like(sci)
    # add a cr
    sci[2, 3] = 10
    crs = flag_crs(sci, err, blot, 1)
    ys, xs = np.where(crs)
    np.testing.assert_equal(ys, 2)
    np.testing.assert_equal(xs, 3)


def test_flag_resampled_crs():
    sci = np.zeros((10, 10), dtype=np.float32)
    err = np.ones_like(sci)
    blot = np.zeros_like(sci)
    # add a cr
    sci[2, 3] = 10

    snr1, snr2 = 5, 4
    scale1, scale2 = 1.2, 0.7
    backg = 0.0
    crs = flag_resampled_crs(sci, err, blot, snr1, snr2, scale1, scale2, backg)
    ys, xs = np.where(crs)
    np.testing.assert_equal(ys, 2)
    np.testing.assert_equal(xs, 3)


def test_gwcs_blot():
    # set up a very simple wcs that scales by 1x
    output_frame = gwcs.Frame2D(name="world")
    forward_transform = models.Scale(1) & models.Scale(1)

    median_data = np.arange(100, dtype=np.float32).reshape((10, 10))
    median_wcs = gwcs.WCS(forward_transform, output_frame=output_frame)
    blot_shape = (5, 5)
    blot_wcs = gwcs.WCS(forward_transform, output_frame=output_frame)
    pix_ratio = 1.0

    blotted = gwcs_blot(median_data, median_wcs, blot_shape, blot_wcs, pix_ratio)
    # since the median data is larger and the wcs are equivalent the blot
    # will window the data to the shape of the blot data
    assert blotted.shape == blot_shape
    np.testing.assert_equal(blotted, median_data[:blot_shape[0], :blot_shape[1]])


@pytest.mark.parametrize('fillval', [0.0, np.nan])
def test_gwcs_blot_fillval(fillval):
    # set up a very simple wcs that scales by 1x
    output_frame = gwcs.Frame2D(name="world")
    forward_transform = models.Scale(1) & models.Scale(1)

    median_shape = (10, 10)
    median_data = np.arange(100, dtype=np.float32).reshape((10, 10))
    median_wcs = gwcs.WCS(forward_transform, output_frame=output_frame)
    blot_shape = (20, 20)
    blot_wcs = gwcs.WCS(forward_transform, output_frame=output_frame)
    pix_ratio = 1.0

    blotted = gwcs_blot(median_data, median_wcs, blot_shape, blot_wcs,
                        pix_ratio, fillval=fillval)

    # since the blot data is larger and the wcs are equivalent the blot
    # will contain the median data + some fill values
    assert blotted.shape == blot_shape
    np.testing.assert_equal(blotted[:median_shape[0], :median_shape[1]], median_data)
    np.testing.assert_equal(blotted[median_shape[0]:, :], fillval)
    np.testing.assert_equal(blotted[:, median_shape[1]:], fillval)


def test_calc_gwcs_pixmap():
    # generate 2 wcses with different scales
    output_frame = gwcs.Frame2D(name="world")
    in_transform = models.Scale(1) & models.Scale(1)
    out_transform = models.Scale(2) & models.Scale(2)
    in_wcs = gwcs.WCS(in_transform, output_frame=output_frame)
    out_wcs = gwcs.WCS(out_transform, output_frame=output_frame)
    in_shape = (3, 4)
    pixmap = calc_gwcs_pixmap(in_wcs, out_wcs, in_shape)
    # we expect given the 2x scale difference to have a pixmap
    # with pixel coordinates / 2
    # use mgrid to generate these coordinates (and reshuffle to match the pixmap)
    expected = np.swapaxes(np.mgrid[:4, :3] / 2., 0, 2)
    np.testing.assert_equal(pixmap, expected)


def test_reproject():
    # generate 2 wcses with different scales
    output_frame = gwcs.Frame2D(name="world")
    wcs1 = gwcs.WCS(models.Scale(1) & models.Scale(1), output_frame=output_frame)
    wcs2 = gwcs.WCS(models.Scale(2) & models.Scale(2), output_frame=output_frame)
    project = reproject(wcs1, wcs2)
    pys, pxs = project(np.array([3]), np.array([1]))
    np.testing.assert_equal(pys, 1.5)
    np.testing.assert_equal(pxs, 0.5)


@pytest.mark.parametrize("shape,kern_size", [
    ([7, 7], [3, 3]),
    ([7, 7], [3, 1]),
    ([7, 7], [1, 3]),
    ([7, 5], [3, 3]),
    ([5, 7], [3, 3]),
    ([42, 42], [7, 7]),
    ([42, 42], [7, 5]),
    ([42, 42], [5, 7]),
    ([42, 7, 5], [3, 3, 3]),
    ([5, 7, 42], [5, 5, 5]),
])
def test_medfilt_against_scipy(shape, kern_size):
    arr = np.arange(np.prod(shape), dtype='uint32').reshape(shape)
    result = medfilt(arr, kern_size)

    # The use of scipy.signal.medfilt is ok here ONLY because the
    # input has no nans. See the medfilt docstring
    expected = scipy.signal.medfilt(arr, kern_size)

    np.testing.assert_allclose(result, expected)


@pytest.mark.parametrize("arr,kern_size,expected", [
    ([2, np.nan, 0], [3], [1, 1, 0]),
    ([np.nan, np.nan, np.nan], [3], [0, np.nan, 0]),
])
def test_medfilt_nan(arr, kern_size, expected):
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message="All-NaN slice",
            category=RuntimeWarning
        )
        result = medfilt(arr, kern_size)
    np.testing.assert_allclose(result, expected)



"""Tests of custom testing infrastructure"""

import pytest
import numpy as np
from stcal.testing_helpers import MemoryThreshold, MemoryThresholdExceeded


def test_memory_threshold():
    with MemoryThreshold("10 KB"):
        buff = np.ones(1000, dtype=np.uint8)


def test_memory_threshold_exceeded():
    with pytest.raises(MemoryThresholdExceeded):
        with MemoryThreshold("500. B"):
            buff = np.ones(10000, dtype=np.uint8)


dqflags = {
    "DO_NOT_USE": None,
    "SATURATED": None,
    "JUMP_DET": None,
    "GOOD": None,
    "NO_GAIN_VALUE": None,
}


def update_dqflags(input_flags):
    dqflags["DO_NOT_USE"] = input_flags["DO_NOT_USE"]
    dqflags["SATURATED"] = input_flags["SATURATED"]
    dqflags["JUMP_DET"] = input_flags["JUMP_DET"]
    dqflags["GOOD"] = input_flags["GOOD"]
    dqflags["NO_GAIN_VALUE"] = input_flags["NO_GAIN_VALUE"]



import importlib
from contextlib import nullcontext

import pytest

try:
    import stdatamodels  # noqa: F401
except ImportError:
    HAS_STDATAMODELS = False
else:
    HAS_STDATAMODELS = True


@pytest.mark.parametrize("name", ["dqflags", "dynamicdq", "basic_utils"])
def test_deprecation(name):
    error = (
        nullcontext()
        if HAS_STDATAMODELS
        else pytest.raises(ImportError, match=f"{name} has been moved to stdatamodels.{name},.*")
    )

    with pytest.warns(DeprecationWarning, match=f"{name} has been moved to stdatamodels.{name},.*"), error:
        importlib.import_module(f"stcal.{name}")



import numpy as np
from scipy import special


class IntegInfo:
    """
    Storage for the integration information for ramp fitting computations.
    """
    def __init__(self, nints, nrows, ncols):
        """
        Initialize output arrays.

        Parameters
        ----------
        nints : int
            The number of integrations in the data.

        nrows : int
            The number of rows in the data.

        ncols : int
            The number of columns in the data.
        """
        dims = (nints, nrows, ncols)
        self.data = np.zeros(shape=dims, dtype=np.float32)

        self.dq = np.zeros(shape=dims, dtype=np.uint32)

        self.var_poisson = np.zeros(shape=dims, dtype=np.float32)
        self.var_rnoise = np.zeros(shape=dims, dtype=np.float32)

        self.err = np.zeros(shape=dims, dtype=np.float32)

    def prepare_info(self):
        """
        Arrange output arrays as a tuple, which the ramp fit step expects.
        """
        return (self.data, self.dq, self.var_poisson, self.var_rnoise, self.err)

    def get_results(self, result, integ, row):
        """
        Capture the ramp fitting computation.

        Parameters
        ----------
        result : RampResult
            Holds computed ramp fitting information.

        integ : int
            The current integration being operated on.

        row : int
            The current row being operated on.
        """
        self.data[integ, row, :] = result.countrate
        self.err[integ, row, :] = result.uncert
        self.var_poisson[integ, row, :] = result.var_poisson
        self.var_rnoise[integ, row, :] = result.var_rnoise


class RampResult:
    def __init__(self):
        """
        Contains the ramp fitting results.
        """
        self.countrate = None
        self.chisq = None
        self.uncert = None
        self.var_poisson = None
        self.var_rnoise = None
        self.weights = None

        self.countrate_two_omit = None
        self.chisq_two_omit = None
        self.uncert_two_omit = None

        self.countrate_one_omit = None
        self.jumpval_one_omit = None
        self.jumpsig_one_omit = None
        self.chisq_one_omit = None
        self.uncert_one_omit = None

    def __repr__(self):
        """
        Return string of information about the class.
        """
        ostring = f"countrate = \n{self.countrate}"
        ostring += f"\nchisq = \n{self.chisq}"
        ostring += f"\nucert = \n{self.uncert}"
        '''
        ostring += f"\nweights = \n{self.weights}"

        ostring += f"\ncountrate_two_omit = \n{self.countrate_two_omit}"
        ostring += f"\nchisq_two_omit = \n{self.chisq_two_omit}"
        ostring += f"\nuncert_two_omit = \n{self.uncert_two_omit}"

        ostring += f"\ncountrate_one_omit = \n{self.countrate_one_omit}"
        ostring += f"\njumpval_one_omit = \n{self.jumpval_one_omit}"
        ostring += f"\njumpsig_one_omit = \n{self.jumpsig_one_omit}"
        ostring += f"\nchisq_one_omit = \n{self.chisq_one_omit}"
        ostring += f"\nuncert_one_omit = \n{self.uncert_one_omit}"
        '''

        return ostring

    def fill_masked_reads(self, diffs2use):
        """
        Mask groups to use for ramp fitting.

        Replace countrates, uncertainties, and chi squared values that
        are NaN because resultant differences were doubly omitted.
        For these cases, revert to the corresponding values in with
        fewer omitted resultant differences to get the correct values
        without double-counting omissions.

        This function replaces the relevant entries of
        self.countrate_two_omit, self.chisq_two_omit,
        self.uncert_two_omit, self.countrate_one_omit, and
        self.chisq_one_omit in place.  It does not return a value.

        Parameters
        ----------
        diffs2use : ndarray
            A 2D array matching self.countrate_one_omit in shape with zero
            for resultant differences that were masked and one for
            differences that were not masked.
        """
        # replace entries that would be nan (from trying to
        # doubly exclude read differences) with the global fits.
        omit = diffs2use == 0
        ones = np.ones(diffs2use.shape)

        self.countrate_one_omit[omit] = (self.countrate * ones)[omit]
        self.chisq_one_omit[omit] = (self.chisq * ones)[omit]
        self.uncert_one_omit[omit] = (self.uncert * ones)[omit]

        omit = diffs2use[1:] == 0

        self.countrate_two_omit[omit] = (self.countrate_one_omit[:-1])[omit]
        self.chisq_two_omit[omit] = (self.chisq_one_omit[:-1])[omit]
        self.uncert_two_omit[omit] = (self.uncert_one_omit[:-1])[omit]

        omit = diffs2use[:-1] == 0

        self.countrate_two_omit[omit] = (self.countrate_one_omit[1:])[omit]
        self.chisq_two_omit[omit] = (self.chisq_one_omit[1:])[omit]
        self.uncert_two_omit[omit] = (self.uncert_one_omit[1:])[omit]


class Covar:
    """
    class Covar holding read and photon noise components of alpha and
    beta and the time intervals between the resultant midpoints
    """
    def __init__(self, readtimes):
        """
        Compute alpha and beta, the diagonal and off-diagonal elements of
        the covariance matrix of the resultant differences, and the time
        intervals between the resultant midpoints.

        Parameters
        ----------
        readtimes : list
            List of values or lists for the times of reads.  If a list of
            lists, times for reads that are averaged together to produce
            a resultant.
        """
        # Equations (4) and (11) in paper 1.
        mean_t, tau, N, delta_t = self._compute_means_and_taus(readtimes)

        self.delta_t = delta_t
        self.mean_t = mean_t
        self.tau = tau
        self.Nreads = N

        # Equations (28) and (29) in paper 1.
        self._compute_alphas_and_betas(mean_t, tau, N, delta_t)

    def _compute_means_and_taus(self, readtimes):
        """
        Computes the means and taus of defined in EQNs 4 and 11 in paper 1.

        Parameters
        ----------
        readtimes : list
            List of values or lists for the times of reads.  If a list of
            lists, times for reads that are averaged together to produce
            a resultant.
        """
        mean_t = []  # mean time of the resultant as defined in the paper
        tau = []  # variance-weighted mean time of the resultant
        N = []  # Number of reads per resultant

        for times in readtimes:
            mean_t += [np.mean(times)]

            if hasattr(times, "__len__"):
                # eqn 11
                N += [len(times)]
                k = np.arange(1, N[-1] + 1)
                if False:
                    tau += [
                        1
                        / N[-1] ** 2
                        * np.sum((2 * N[-1] + 1 - 2 * k) * np.array(times))
                    ]
                    # tau += [(np.sum((2*N[-1] + 1 - 2*k)*np.array(times))) / N[-1]**2]
                else:
                    length = N[-1]
                    tmp0 = (2 * length + 1) - (2 * k)
                    sm = np.sum(tmp0 * np.array(times))
                    tmp = sm / length**2
                    tau.append(tmp)
            else:
                tau += [times]
                N += [1]

        # readtimes is a list of lists, so mean_t is the list of each
        # mean of each list.
        mean_t = np.array(mean_t)
        tau = np.array(tau)
        N = np.array(N)
        delta_t = mean_t[1:] - mean_t[:-1]

        return mean_t, tau, N, delta_t

    def _compute_alphas_and_betas(self, mean_t, tau, N, delta_t):
        """
        Computes the means and taus defined in EQNs 28 and 29 in paper 1.

        Parameters
        ----------
        mean_t : ndarray
            The means of the reads for each group.

        tau : ndarray
            Intermediate computation.

        N : ndarray
            The number of reads in each group.

        delta_t : ndarray
            The group differences of integration ramps.
        """
        self.alpha_readnoise = (1 / N[:-1] + 1 / N[1:]) / delta_t**2
        self.beta_readnoise = -1 / (N[1:-1] * delta_t[1:] * delta_t[:-1])

        self.alpha_phnoise = (tau[:-1] + tau[1:] - 2 * mean_t[:-1]) / delta_t**2
        self.beta_phnoise = (mean_t[1:-1] - tau[1:-1]) / (delta_t[1:] * delta_t[:-1])

    def calc_bias(self, countrates, sig, cvec, da=1e-7):
        """
        Calculate the bias in the best-fit count rate from estimating the covariance matrix.

        Section 5 of paper 1.

        Arguments:
        Parameters
        ----------
        countrates : ndarray
            Array of count rates at which the bias is desired.

        sig : float
            Single read noise.

        cvec : ndarray
            Weight vector on resultant differences for initial estimation
            of count rate for the covariance matrix. Will be renormalized
            inside this function.

        da : float
            Fraction of the count rate plus sig**2 to use for finite difference
            estimate of the derivative.  Optional parameter.  Default 1e-7.

        Returns
        -------
        bias : ndarray
            Bias of the best-fit count rate from using cvec plus the observed
            resultants to estimate the covariance matrix.
        """
        alpha = countrates[np.newaxis, :] * self.alpha_phnoise[:, np.newaxis]
        alpha += sig**2 * self.alpha_readnoise[:, np.newaxis]
        beta = countrates[np.newaxis, :] * self.beta_phnoise[:, np.newaxis]
        beta += sig**2 * self.beta_readnoise[:, np.newaxis]

        # we only want the weights; it doesn't matter what the count rates are.
        n = alpha.shape[0]
        z = np.zeros((len(cvec), len(countrates)))
        result_low_a = fit_ramps(z, self, sig, countrateguess=countrates)

        # try to avoid problems with roundoff error
        da_incr = da * (countrates[np.newaxis, :] + sig**2)

        dalpha = da_incr * self.alpha_phnoise[:, np.newaxis]
        dbeta = da_incr * self.beta_phnoise[:, np.newaxis]
        result_high_a = fit_ramps(z, self, sig, countrateguess=countrates + da_incr)
        # finite difference approximation to dw/da

        dw_da = (result_high_a.weights - result_low_a.weights) / da_incr

        bias = np.zeros(len(countrates))
        c = cvec / np.sum(cvec)

        for i in range(len(countrates)):

            C = np.zeros((n, n))
            for j in range(n):
                C[j, j] = alpha[j, i]
            for j in range(n - 1):
                C[j + 1, j] = C[j, j + 1] = beta[j, i]

            bias[i] = np.linalg.multi_dot([c[np.newaxis, :], C, dw_da[:, i]])

            sig_a = np.sqrt(
                np.linalg.multi_dot([c[np.newaxis, :], C, c[:, np.newaxis]])
            )
            bias[i] *= 0.5 * (1 + special.erf(countrates[i] / sig_a / 2**0.5))

        return bias



import numpy as np


def linearity_correction(data, gdq, pdq, lin_coeffs, lin_dq, dqflags, zframe=None):
    """
    Apply linearity correction to individual groups in `data` to pixels that
    haven't already been flagged as saturated.

    Pixels having at least one correction coefficient equal to NaN will not
    have the linearity correction applied and the DQ flag `NO_LIN_CORR` is
    added to the science exposure PIXELDQ array. Pixels that have the
    `NO_LIN_CORR` flag set in the DQ array of the linearity reference file will
    not have the correction applied and the “NO_LIN_CORR” flag is added to the
    science exposure `PIXELDQ` array. Pixel values that have the `SATURATED`
    flag set in a particular group of the science exposure GROUPDQ array will
    not have the linearity correction applied to that group. Any groups for
    that pixel that are not flagged as saturated will be corrected.

    The corrected data and updated `PIXELDQ` arrays are returned.
    Reference:
    http://www.stsci.edu/hst/wfc3/documents/handbooks/currentDHB/wfc3_Ch606.html
    WFC3 Data Handbook, Chapter 6, WFC3-IR Error Sources, 6.5 Detector
    Nonlinearity issues, 2.1 May 2011

    Parameters
    ----------
    data : ndarray
        The 4D array for the SCI data.

    gdq : ndarray
        The 4D group dq array.

    pdq : ndarray
        The 2D pixel dq array.

    lin_coeffs : ndarray.
        The 3D array containing pixel-by-pixel linearity coefficient
        values for each term in the polynomial fit.

    dq_flags : dict
        A dictionary with at least the following keywords:
        SATURATED, NO_LIN_CORR.

    zframe : ndarray or None
        The ZEROFRAME array with dimensions (nints, nrows, ncols) or None

    Returns
    -------
    data : ndarray
        The 4D linearly corrected SCI data.

    new_pdq : ndarray
        The updated 2D pixel data quality flags.

    zframe : ndarray or None
        If not None, the 3D linearly corrected ZEROFRAME.
    """
    if zframe is not None:
        # Save off data that gets transformed during linearity processing.
        zlin_coeffs = lin_coeffs.copy()
        zlin_dq = lin_dq.copy()

    # Do linear correction on SCI data
    data, new_pdq = linearity_correction_branch(data, gdq, pdq, lin_coeffs, lin_dq, dqflags, False)

    zdata = None  # zframe needs to be returned, so initialize it to None.
    if zframe is not None:
        # Do linear correction on ZEROFRAME
        # As ZEROFRAME gets processed through the pipeline, since it has no
        # corresponding DQ array, when data is found to be bad, the data is
        # set to ZERO.  Since zero ZEROFRAME values indicates bad data,
        # remember where this happens.  Make a dummy ZEROFRAME DQ array and
        # mark zeroed data as saturated.
        wh_zero = np.where(zframe[:, :, :] == 0.0)
        zdq = np.zeros(zframe.shape, dtype=gdq.dtype)
        zdq[zframe == 0.0] = dqflags["SATURATED"]
        zpdq = np.zeros(zframe.shape[-2:], dtype=pdq.dtype)

        # Linearly correct ZEROFRAME
        zdata, _ = linearity_correction_branch(zframe, zdq, zpdq, zlin_coeffs, zlin_dq, dqflags, True)

        # Ensure bad data remains bad.
        zdata[wh_zero] = 0.0

    return data, new_pdq, zdata


def linearity_correction_branch(data, gdq, pdq, lin_coeffs, lin_dq, dqflags, zframe):
    """
    Parameters
    ----------
    data : `np.array`
        The data to be linearly corrected.

    gdq : `np.array`
        Group dq array.

    pdq : `np.array`
        Pixel dq array.

    lin_coeffs : `np.array`
        Array containing pixel-by-pixel linearity coefficient values
        for each term in the polynomial fit.

    dq_flags : dict
        A dictionary with at least the following keywords:
        SATURATED, NO_LIN_CORR.

    zframe : ndarray or None
        The ZEROFRAME array with dimensions (nints, nrows, ncols) or None

    Returns
    -------
    data : 3D array
        Updated array of correction coefficients in reference file.

    new_pdq : `np.array`
        Updated pixel data quality flags.
    """
    # Retrieve the ramp data cube characteristics
    if zframe:
        nints, nrows, ncols = data.shape
    else:
        nints, ngroups, nrows, ncols = data.shape

    # Number of coeffs is equal to the number of planes in coeff cube
    ncoeffs = lin_coeffs.shape[0]

    # Combine the DQ arrays using bitwise_or
    new_pdq = np.bitwise_or(pdq, lin_dq)

    # Check for NO_LIN_CORR flags in the DQ extension of the ref file
    lin_coeffs = correct_for_flag(lin_coeffs, lin_dq, dqflags)

    # Check for NaNs in the COEFFS extension of the ref file
    lin_coeffs, new_pdq = correct_for_NaN(lin_coeffs, new_pdq, dqflags)

    # Check when all the Linearity COEFFS = 0. Set DQ flag to NO_LIN_COEFF
    lin_coeffs, new_pdq = correct_for_zero(lin_coeffs, new_pdq, dqflags)

    # Apply the linearity correction one integration at a time.
    for ints in range(nints):
        if not zframe:
            # Apply the linearity correction one group at a time
            for plane in range(ngroups):
                dataplane = data[ints, plane]
                gdqplane = gdq[ints, plane]
                linear_correct_plane(dataplane, gdqplane, lin_coeffs, ncoeffs, dqflags)

        else:
            # ZEROFRAME processing
            dataplane = data[ints]
            gdqplane = gdq[ints]
            linear_correct_plane(dataplane, gdqplane, lin_coeffs, ncoeffs, dqflags)

    return data, new_pdq


def linear_correct_plane(dataplane, gdqplane, lin_coeffs, ncoeffs, dqflags):
    """
    dataplane : ndarray
        The 2D array of the frame/group plane of pixels to linearly correct.

    gdqplane : ndarray
        The DQ flags for the plane.

    lin_coeffs : ndarray
        The linearity coefficients to apply to the pixels.

    ncoeffs : int
        The number of linearity coefficients.

    dqflags : dict
        The dictionary of DQ flags.
    """
    # Accumulate the polynomial terms into the corrected counts
    scorr = lin_coeffs[ncoeffs - 1] * dataplane
    for j in range(ncoeffs - 2, 0, -1):
        scorr = (scorr + lin_coeffs[j]) * dataplane
    scorr = lin_coeffs[0] + scorr
    # Only use the corrected signal where the original signal value
    # has not been flagged by the saturation step.
    # Otherwise use the original signal.
    dataplane[:, :] = np.where(np.bitwise_and(gdqplane[:, :], dqflags["SATURATED"]), dataplane[:, :], scorr)


def correct_for_NaN(lin_coeffs, pixeldq, dqflags):
    """
    Check for NaNs in the COEFFS extension of the ref file in case there are
    pixels that should have been (but were not) flagged there as NO_LIN_CORR
    (linearity correction not determined for pixel).

    For such pixels, update the
    coefficients so that there is effectively no correction, and flag their
    pixeldq values in place as NO_LIN_CORR in the step output.

    Parameters
    ----------
    lin_coeffs: 3D array
        array of correction coefficients in reference file

    input: data model object
        science data model to be corrected in place

    Returns
    -------
    lin_coeffs: 3D array
        updated array of correction coefficients in reference file
    """
    wh_nan = np.where(np.isnan(lin_coeffs))
    znan, ynan, xnan = wh_nan[0], wh_nan[1], wh_nan[2]
    num_nan = 0

    nan_array = np.zeros((lin_coeffs.shape[1], lin_coeffs.shape[2]), dtype=np.uint32)

    # If there are NaNs as the correction coefficients, update those
    # coefficients so that those SCI values will be unchanged.
    if len(znan) > 0:
        ben_cor = ben_coeffs(lin_coeffs)  # get benign coefficients
        num_nan = len(znan)

        for ii in range(num_nan):
            lin_coeffs[:, ynan[ii], xnan[ii]] = ben_cor
            nan_array[ynan[ii], xnan[ii]] = dqflags["NO_LIN_CORR"]

        # Include these pixels in the output pixeldq
        pixeldq = np.bitwise_or(pixeldq, nan_array)

    return lin_coeffs, pixeldq


def correct_for_zero(lin_coeffs, pixeldq, dqflags):
    """
    Check when the linear term in the linearity coefficients is zero.  For such pixels, update the
    coefficients so that there is effectively no correction, and flag their
    pixeldq values in place as NO_LIN_CORR in the step output.

    Parameters
    ----------
    lin_coeffs: 3D array
        array of correction coefficients in reference file

    input: data model object
        science data model to be corrected in place

    Returns
    -------
    lin_coeffs: 3D array
        updated array of correction coefficients in reference file
    """
    # The critical coefficient that should not be zero is the linear term other terms are fine to be zero
    linear_term = lin_coeffs[1, :, :]
    wh_zero = np.where(linear_term == 0)
    yzero, xzero = wh_zero[0], wh_zero[1]
    num_zero = 0
    lin_dq_array = np.zeros((lin_coeffs.shape[1], lin_coeffs.shape[2]), dtype=np.uint32)

    # If there are linearity linear term equal to zero,
    # update the coefficients so the SCI values will be unchanged.
    if len(yzero) > 0:
        ben_cor = ben_coeffs(lin_coeffs)  # get benign coefficients
        num_zero = len(yzero)

        for ii in range(num_zero):
            lin_coeffs[:, yzero[ii], xzero[ii]] = ben_cor
            lin_dq_array[yzero[ii], xzero[ii]] = dqflags["NO_LIN_CORR"]

        # Include these pixels in the output pixeldq
        pixeldq = np.bitwise_or(pixeldq, lin_dq_array)

    return lin_coeffs, pixeldq


def correct_for_flag(lin_coeffs, lin_dq, dqflags):
    """
    Short Summary
    -------------
    Check for pixels that are flagged as NO_LIN_CORR
    ('No linearity correction available') in the DQ extension of the ref data.
    For such pixels, update the coefficients so that there is effectively no
    correction. Because these are already flagged in the ref file, they will
    also be flagged in the output dq.

    Parameters
    ----------
    lin_coeffs: 3D array
        array of correction coefficients in reference file

    lin_dq: 2D array
        array of data quality flags in reference file

    Returns
    -------
    lin_coeffs: 3D array
        updated array of correction coefficients in reference file
    """
    wh_flag = np.bitwise_and(lin_dq, dqflags["NO_LIN_CORR"])
    num_flag = len(np.where(wh_flag > 0)[0])

    wh_lin = np.where(wh_flag == dqflags["NO_LIN_CORR"])
    yf, xf = wh_lin[0], wh_lin[1]

    # If there are pixels flagged as 'NO_LIN_CORR', update the corresponding
    #     coefficients so that those SCI values will be unchanged.
    if num_flag > 0:
        ben_cor = ben_coeffs(lin_coeffs)  # get benign coefficients

        for ii in range(num_flag):
            lin_coeffs[:, yf[ii], xf[ii]] = ben_cor

    return lin_coeffs


def ben_coeffs(lin_coeffs):
    """
    Short Summary
    -------------
    For pixels having at least 1 NaN coefficient, reset the coefficients to be
    benign, which will effectively leave the SCI values unaffected.

    Parameters
    ----------
    lin_coeffs: 3D array
        array of correction coefficients in reference file

    Returns
    -------
    ben_cor: 1D array
        benign coefficients - all ben_cor[:] = 0.0 except ben_cor[1] = 1.0
    """
    ben_cor = np.zeros(lin_coeffs.shape[0])
    ben_cor[1] = 1.0

    return ben_cor



#! /usr/bin/env python

import logging
import multiprocessing
import time
import scipy
import sys
import warnings

from multiprocessing import cpu_count
from pprint import pprint

import numpy as np

from . import ramp_fit_class, utils
from .likely_algo_classes import IntegInfo, RampResult, Covar


DELIM = "=" * 80
SQRT2 = np.sqrt(2)
LIKELY_MIN_NGROUPS = 4

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


def likely_ramp_fit(ramp_data, readnoise_2d, gain_2d):
    """
    Invoke ramp fitting using the likelihood algorithm.

    Parameters
    ----------
    ramp_data : RampData
        Input data necessary for computing ramp fitting.

    readnoise_2d : ndarray
        readnoise for all pixels

    gain_2d : ndarray
        gain for all pixels

    Returns
    -------
    image_info : tuple
        The tuple of computed ramp fitting arrays.

    integ_info : tuple
        The tuple of computed integration fitting arrays.

    opt_info : tuple
        The tuple of computed optional results arrays for fitting.
    """
    image_info, integ_info, opt_info = None, None, None

    nints, ngroups, nrows, ncols = ramp_data.data.shape

    if ngroups < LIKELY_MIN_NGROUPS:
        raise ValueError("Likelihood fit requires at least 4 groups.")


    remove_jump_detection_flags(ramp_data)

    readtimes = get_readtimes(ramp_data)

    covar = Covar(readtimes)
    integ_class = IntegInfo(nints, nrows, ncols)

    readnoise_2d = readnoise_2d / SQRT2

    for integ in range(nints):
        data = ramp_data.data[integ, :, :, :]
        gdq = ramp_data.groupdq[integ, :, :, :].copy()
        pdq = ramp_data.pixeldq[:, :].copy()

        # Eqn (5)
        diff = (data[1:] - data[:-1]) / covar.delta_t[:, np.newaxis, np.newaxis]
        alldiffs2use = np.ones(diff.shape, np.uint8)

        for row in range(nrows):
            d2use = determine_diffs2use(ramp_data, integ, row, diff)
            d2use_copy = d2use.copy()  # Use to flag jumps
            if ramp_data.rejection_threshold is not None:
                threshold_one_omit = ramp_data.rejection_threshold**2
                pval = scipy.special.erfc(ramp_data.rejection_threshold/SQRT2)
                threshold_two_omit = scipy.stats.chi2.isf(pval, 2)
                if np.isinf(threshold_two_omit):
                    threshold_two_omit = threshold_one_omit + 10
                d2use, countrates = mask_jumps(
                    diff[:, row], covar, readnoise_2d[row], gain_2d[row],
                    threshold_one_omit=threshold_one_omit,
                    threshold_two_omit=threshold_two_omit,
                    diffs2use=d2use
                )
            else:
                d2use, countrates = mask_jumps(
                    diff[:, row], covar, readnoise_2d[row], gain_2d[row],
                    diffs2use=d2use
                )

            # Set jump detection flags
            jump_locs = d2use_copy ^ d2use
            jump_locs[jump_locs > 0] = ramp_data.flags_jump_det
            gdq[1:, row, :] |= jump_locs

            alldiffs2use[:, row] = d2use

            rateguess = countrates * (countrates > 0) + ramp_data.average_dark_current[row, :]
            result = fit_ramps(
                diff[:, row],
                covar,
                gain_2d[row],
                readnoise_2d[row],
                diffs2use=d2use,
                count_rate_guess=rateguess,
            )
            integ_class.get_results(result, integ, row)

        pdq = utils.dq_compress_sect(ramp_data, integ, gdq, pdq)
        integ_class.dq[integ, :, :] = pdq

        del gdq

    integ_info = integ_class.prepare_info()
    image_info = compute_image_info(integ_class, ramp_data)

    return image_info, integ_info, opt_info


def mask_jumps(
    diffs,
    covar,
    rnoise,
    gain,
    threshold_one_omit=20.25,
    threshold_two_omit=23.8,
    diffs2use=None,
):

    """
    Implements a likelihood-based, iterative jump detection algorithm.

    Parameters
    ----------
    diffs : ndarray
        The group differences of the data array for a given integration and row
        (ngroups-1, ncols).

    covar : Covar
        The class instance that computes and contains the covariance matrix info.

    rnoise : ndarray
        The read noise (ncols,)

    gain : ndarray
        The gain (ncols,)

    threshold_one_omit : float
        Minimum chisq improvement to exclude a single resultant difference.
        Default: 20.25.

    threshold_two_omit : float
        Minimum chisq improvement to exclude  two sequential resultant differences.
        Default 23.8.

    diffs2use : ndarray
        A boolean array definined the segmented ramps for each pixel in a row.
        (ngroups-1, ncols)

    Returns
    -------
    dffs2use : ndarray
        A boolean array definined the segmented ramps for each pixel in a row.
        (ngroups-1, ncols)

    countrates : ndarray
        Count rate estimates used to estimate the covariance matrix.
        Optional, default is None.

    """
    # Force a copy of the input array for more efficient memory access.
    # loc_diff = diffs * 1
    loc_diff = diffs

    # We can use one-omit searches only where the reads immediately
    # preceding and following have just one read.  If a readout
    # pattern has more than one read per resultant but significant
    # gaps between resultants then a one-omit search might still be a
    # good idea even with multiple-read resultants.
    one_omit_ok = covar.Nreads[1:] * covar.Nreads[:-1] >= 1
    one_omit_ok[0] = one_omit_ok[-1] = True

    # Other than that, we need to omit two.  If a resultant has more
    # than two reads, we need to omit both differences containing it
    # (one pair of omissions in the differences).
    two_omit_ok = covar.Nreads[1:-1] > 1

    # This is the array to return: one for resultant differences to
    # use, zero for resultant differences to ignore.
    if diffs2use is None:
        diffs2use = np.ones(loc_diff.shape, np.uint8)

    # We need to estimate the covariance matrix.  I'll use the median
    # here for now to limit problems with the count rate in reads with
    # jumps (which is what we are looking for) since we'll be using
    # likelihoods and chi squared; getting the covariance matrix
    # reasonably close to correct is important.
    count_rate_guess = np.median(loc_diff, axis=0)[np.newaxis, :]

    count_rate_guess *= count_rate_guess > 0

    # boolean arrays to be used later
    recheck = np.ones(loc_diff.shape[1]) == 1
    dropped = np.ones(loc_diff.shape[1]) == 0

    for j in range(loc_diff.shape[0]):
        # No need for indexing on the first pass.
        if j == 0:
            result = fit_ramps(
                loc_diff,
                covar,
                gain,
                rnoise,
                count_rate_guess=count_rate_guess,
                diffs2use=diffs2use,
                detect_jumps=True,
            )
            # Also save the count rates so that we can use them later
            # for debiasing.
            countrate = result.countrate * 1.0
        else:
            result = fit_ramps(
                loc_diff[:, recheck],
                covar,
                gain[recheck],
                rnoise[recheck],
                count_rate_guess=count_rate_guess[:, recheck],
                diffs2use=diffs2use[:, recheck],
                detect_jumps=True,
            )

        # Chi squared improvements
        dchisq_two = result.chisq - result.chisq_two_omit
        dchisq_one = result.chisq - result.chisq_one_omit

        # We want the largest chi squared difference
        best_dchisq_one = np.amax(dchisq_one * one_omit_ok[:, np.newaxis], axis=0)
        best_dchisq_two = np.amax(
            dchisq_two * two_omit_ok[:, np.newaxis], axis=0
        )

        # Is the best improvement from dropping one resultant
        # difference or two?  Two drops will always offer more
        # improvement than one so penalize them by the respective
        # thresholds.  Then find the chi squared improvement
        # corresponding to dropping either one or two reads, whichever
        # is better, if either exceeded the threshold.
        onedropbetter = (
            best_dchisq_one - threshold_one_omit > best_dchisq_two - threshold_two_omit
        )

        best_dchisq = (
            best_dchisq_one * (best_dchisq_one > threshold_one_omit) * onedropbetter
        )
        best_dchisq += (
            best_dchisq_two * (best_dchisq_two > threshold_two_omit) * (~onedropbetter)
        )

        # If nothing exceeded the threshold set the improvement to
        # NaN so that dchisq==best_dchisq is guaranteed to be False.
        best_dchisq[best_dchisq == 0] = np.nan

        # Now make the masks for which resultant difference(s) to
        # drop, count the number of ramps affected, and drop them.
        # If no ramps were affected break the loop.
        dropone = dchisq_one == best_dchisq
        droptwo = dchisq_two == best_dchisq

        drop = np.any([np.sum(dropone, axis=0), np.sum(droptwo, axis=0)], axis=0)

        if np.sum(drop) == 0:
            break

        # Store the updated counts with omitted reads
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            new_cts = np.zeros(np.sum(recheck))
            i_d1 = np.sum(dropone, axis=0) > 0
            new_cts[i_d1] = np.sum(result.countrate_one_omit * dropone, axis=0)[i_d1]
            i_d2 = np.sum(droptwo, axis=0) > 0
            new_cts[i_d2] = np.sum(result.countrate_two_omit * droptwo, axis=0)[i_d2]

        # zero out count rates with drops and add their new values back in
        countrate[recheck] *= drop == 0
        countrate[recheck] += new_cts

        # Drop the read (set diffs2use=0) if the boolean array is True.
        diffs2use[:, recheck] *= ~dropone
        diffs2use[:-1, recheck] *= ~droptwo
        diffs2use[1:, recheck] *= ~droptwo

        # No need to repeat this on the entire ramp, only re-search
        # ramps that had a resultant difference dropped this time.
        dropped[:] = False
        dropped[recheck] = drop
        recheck[:] = dropped

        # Do not try to search for bad resultants if we have already
        # given up on all but one, two, or three resultant differences
        # in the ramp.  If there are only two left we have no way of
        # choosing which one is "good".  If there are three left we
        # run into trouble in case we need to discard two.
        recheck[np.sum(diffs2use, axis=0) <= 3] = False

    return diffs2use, countrate


def get_readtimes(ramp_data):
    """
    Get the read times needed to compute the covariance matrices.

    If there is already a read_pattern in the ramp_data class, then just get it.
    If not, then one needs to be constructed.  If one needs to be constructed it
    is assumed the groups are evenly spaced in time, as are the frames that make
    up the group.  If each group has only one frame and no group gap, then a list
    of the group times is returned.  If nframes > 0, then a list of lists of each
    frame time in each group is returned with the assumption:
        group_time = (nframes + groupgap) * frame_time

    Parameters
    ----------
    ramp_data : RampData
        Input data necessary for computing ramp fitting.

    Returns
    -------
    readtimes : list
        A list of frame times for each frame used in the computation of the ramp.
    """
    if ramp_data.read_pattern is not None:
        return ramp_data.read_pattern

    ngroups = ramp_data.data.shape[1]
    tot_frames = ramp_data.nframes + ramp_data.groupgap
    tot_nreads = np.arange(1, ramp_data.nframes + 1)
    rtimes = [
        (tot_nreads + k * tot_frames) * ramp_data.frame_time for k in range(ngroups)
    ]

    return rtimes


def compute_image_info(integ_class, ramp_data):
    """
    Combine all integrations into a single image of rates,
    variances, and DQ flags.

    Parameters
    ----------
    integ_class : IntegInfo
        Contains the rateints product calculations.

    ramp_data : RampData
        Input data necessary for computing ramp fitting.

    Returns
    -------
    image_info : tuple
        The list of arrays for the rate product.
    """
    if integ_class.data.shape[0] == 1:
        data = integ_class.data[0, :, :]
        dq = integ_class.dq[0, :, :]
        var_p = integ_class.var_poisson[0, :, :]
        var_r = integ_class.var_rnoise[0, :, :]
        var_e = integ_class.err[0, :, :]
        return (data, dq, var_p, var_r, var_e)

    dq = utils.dq_compress_final(integ_class.dq, ramp_data)

    slope = np.nanmedian(integ_class.data, axis=0)
    for _ in range(2):
        rate_scale = slope[np.newaxis, :] / integ_class.data
        rate_scale[(~np.isfinite(rate_scale)) | (rate_scale < 0)] = 0
        all_var_p = integ_class.var_poisson * rate_scale
        weight = 1/(all_var_p + integ_class.var_rnoise)
        weight /= np.nansum(weight, axis=0)[np.newaxis, :]
        tmp_slope = integ_class.data * weight
        all_nan = np.all(np.isnan(tmp_slope), axis=0)
        slope = np.nansum(tmp_slope, axis=0)
        slope[all_nan] = np.nan

    tmp_v = all_var_p * weight**2
    all_nan = np.all(np.isnan(tmp_v), axis=0)
    var_p = np.nansum(tmp_v, axis=0)
    var_p[all_nan] = np.nan

    tmp_v = integ_class.var_rnoise * weight**2
    all_nan = np.all(np.isnan(tmp_v), axis=0)
    var_r = np.nansum(tmp_v, axis=0)
    var_r[all_nan] = np.nan

    err = np.sqrt(var_p + var_r)

    return (slope, dq, var_p, var_r, err)


def remove_jump_detection_flags(ramp_data):
    """
    Remove the JUMP_DET flag from the group DQ array

    Parameters
    ----------
    ramp_data : RampData
        Input data necessary for computing ramp fitting.
    """
    jump = ramp_data.flags_jump_det
    gdq = ramp_data.groupdq
    wh_jump = np.where(np.bitwise_and(gdq.astype(np.uint32), jump))
    gdq[wh_jump] -= jump
    ramp_data.groupdq = gdq


def determine_diffs2use(ramp_data, integ, row, diffs):
    """
    Compute the diffs2use mask based on DQ flags of a row.

    Parameters
    ----------
    ramp_data : RampData
        Input data necessary for computing ramp fitting.

    integ : int
        The current integration being processed.

    row : int
        The current row being processed.

    diffs : ndarray
        The group differences of the data array for a given integration and row
        (ngroups-1, ncols).

    Returns
    -------
    d2use : ndarray
        A boolean array definined the segmented ramps for each pixel in a row.
        (ngroups-1, ncols)
    """
    # import ipdb; ipdb.set_trace()
    _, ngroups, _, ncols = ramp_data.data.shape
    dq = np.zeros(shape=(ngroups, ncols), dtype=np.uint8)
    dq[:, :] = ramp_data.groupdq[integ, :, row, :]
    d2use_tmp = np.ones(shape=diffs.shape, dtype=np.uint8)
    d2use = d2use_tmp[:, row]

    d2use[dq[1:, :] != 0] = 0
    d2use[dq[:-1, :] != 0] = 0

    return d2use


def initial_count_rate_guess(diffs, diffs2use):
    """
    Compute the initial count rate.

    Parameters
    ----------
    diffs : ndarray
        The group differences of the data (ngroups-1, nrows, ncols).

    diffs2use : ndarray
        Boolean mask determining with group differences to use (ngroups-1, ncols).

    Returns
    -------
    count_rate_guess : ndarray
        The initial count rate.
    """
    # initial guess for count rate is the average of the unmasked
    # group differences unless otherwise specified.
    num = np.sum((diffs * diffs2use), axis=0)
    den = np.sum(diffs2use, axis=0)

    count_rate_guess = num / den
    count_rate_guess *= count_rate_guess > 0

    return count_rate_guess


# RAMP FITTING BEGIN
def fit_ramps(
    diffs,
    covar,
    gain,
    rnoise,  # Referred to as 'sig' in fitramp repo
    count_rate_guess=None,
    diffs2use=None,
    detect_jumps=False,
    rescale=True,
    dn_scale=10.0,
):
    """
    Fit ramps on a row of data.

    Fits ramps to read differences using the covariance matrix for
    the read differences as given by the diagonal elements and the
    off-diagonal elements.

    Parameters
    ----------
    diffs : ndarray
        The group differences of the data (ngroups-1, nrows, ncols).

    covar : Covar
        The class instance that computes and contains the covariance matrix info.

    gain : ndarray
        The gain (ncols,)

    rnoise : ndarray
        The read noise (ncols,)

    count_rate_guess : ndarray
        Count rate estimates used to estimate the covariance matrix.
        Optional, default is None.

    diffs2use : ndarray
        Boolean mask determining with group differences to use (ngroups-1, ncols).
        Optional, default is None, which results in a mask of all 1's.

    detect_jumps : boolean
        Run jump detection.
        Optional, default is False.

    rescale : boolean
        Scale the covariance matrix internally to avoid possible
        overflow/underflow problems for long ramps.
        Optional, default is True.

    dn_scale : float
        Scaling factor.

    Returns
    -------
    result : RampResult
        Holds computed ramp fitting information.  XXX - rename
    """
    if diffs2use is None:
        # Use all diffs
        diffs2use = np.ones(diffs.shape, np.uint8)

    # diffs is (ngroups, ncols) of the current row
    if count_rate_guess is None:
        count_rate_guess = initial_count_rate_guess(covar, diffs, diffs2use)

    alpha_tuple, beta_tuple, scale = compute_alphas_betas(
        count_rate_guess, gain, rnoise, covar, rescale, diffs, dn_scale
    )
    alpha, alpha_phnoise, alpha_readnoise = alpha_tuple
    beta, beta_phnoise, beta_readnoise = beta_tuple

    ndiffs, npix = diffs.shape

    # Mask group differences that should be ignored.  This is half
    # of what we need to do to mask these group differences; the
    # rest comes later.
    diff_mask = diffs * diffs2use
    beta = beta * diffs2use[1:] * diffs2use[:-1]

    # All definitions and formulas here are in the paper.
    # --- Till line 237: Paper 1 section 4
    theta = compute_thetas(ndiffs, npix, alpha, beta)  # EQNs 38-40
    phi = compute_phis(ndiffs, npix, alpha, beta)  # EQNs 41-43

    sgn = np.ones((ndiffs, npix))
    sgn[::2] = -1

    Phi = compute_Phis(ndiffs, npix, beta, phi, sgn)  # EQN 46
    PhiD = compute_PhiDs(ndiffs, npix, beta, phi, sgn, diff_mask)  # EQN ??
    Theta = compute_Thetas(ndiffs, npix, beta, theta, sgn)  # EQN 47
    ThetaD = compute_ThetaDs(ndiffs, npix, beta, theta, sgn, diff_mask)  # EQN 48

    dB, dC, A, B, C = matrix_computations(
        ndiffs,
        npix,
        sgn,
        diff_mask,
        diffs2use,
        beta,
        phi,
        Phi,
        PhiD,
        theta,
        Theta,
        ThetaD,
    )

    result = get_ramp_result(
        dC,
        A,
        B,
        C,
        scale,
        alpha_phnoise,
        alpha_readnoise,
        beta_phnoise,
        beta_readnoise,
    )

    # --- Beginning at line 250: Paper 1 section 4

    if detect_jumps:
        # result = compute_jump_detects(
        #     result, ndiffs, diffs2use, dC, dB, A, B, C, scale, beta, phi, theta, covar
        # )
        result = compute_jump_detects(
            result, ndiffs, diffs2use, dC, dB, A, B, C, scale, beta, phi, theta
        )

    return result


# RAMP FITTING END


def compute_jump_detects(
    result, ndiffs, diffs2use, dC, dB, A, B, C, scale, beta, phi, theta
):
    """
    Detect jumps in ramps.

    The code below computes the best chi squared, best-fit slope,
    and its uncertainty leaving out each group difference in
    turn.  There are ndiffs possible differences that can be
    omitted.

    Then do it omitting two consecutive reads.  There are ndiffs-1
    possible pairs of adjacent reads that can be omitted.

    Paper II, sections 3.1 and 3.2

    Parameters
    ----------
    result : RampResult
        The results of the ramp fitting for a given row of pixels in an integration.

    ndiffs :  int
        Number of differences.

    diffs2use : ndarray
        Boolean mask determining with group differences to use (ngroups-1, ncols).

    dC : ndarray
        Intermediate computation.

    dB : ndarray
        Intermediate computation.

    A : ndarray
        Intermediate computation.

    B : ndarray
        Intermediate computation.

    C : ndarray
        Intermediate computation.

    scale : ndarray or integer
        Overflow/underflow prevention scale.

    beta : ndarray
        Off diagonal of covariance matrix.

    phi : ndarray
        Intermediate computation.

    theta : ndarray
        Intermediate computation.

    covar : Covar
        The class instance that computes and contains the covariance matrix info.


    Returns
    -------
    result : RampResult
        The results of the ramp fitting for a given row of pixels in an integration.
    """
    # Diagonal elements of the inverse covariance matrix
    Cinv_diag = theta[:-1] * phi[1:] / theta[ndiffs]
    Cinv_diag *= diffs2use

    # Off-diagonal elements of the inverse covariance matrix
    # one spot above and below for the case of two adjacent
    # differences to be masked
    Cinv_offdiag = -beta * theta[:-2] * phi[2:] / theta[ndiffs]

    # Equations in the paper: best-fit a, b
    #
    # Catch warnings in case there are masked group
    # differences, since these will be overwritten later.  No need
    # to warn about division by zero here.
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        a = (Cinv_diag * B - dB * dC) / (C * Cinv_diag - dC**2)
        b = (dB - a * dC) / Cinv_diag

        result.countrate_one_omit = a
        result.jumpval_one_omit = b

        # Use the best-fit a, b to get chi squared
        result.chisq_one_omit = (
            A
            + a**2 * C
            - 2 * a * B
            + b**2 * Cinv_diag
            - 2 * b * dB
            + 2 * a * b * dC
        )

        # invert the covariance matrix of a, b to get the uncertainty on a
        result.uncert_one_omit = np.sqrt(Cinv_diag / (C * Cinv_diag - dC**2))
        result.jumpsig_one_omit = np.sqrt(C / (C * Cinv_diag - dC**2))

        result.chisq_one_omit /= scale
        result.uncert_one_omit *= np.sqrt(scale)
        result.jumpsig_one_omit *= np.sqrt(scale)

    # Now for two omissions in a row.  This is more work.  Again,
    # all equations are in the paper.  I first define three
    # factors that will be used more than once to save a bit of
    # computational effort.
    cpj_fac = dC[:-1] ** 2 - C * Cinv_diag[:-1]
    cjck_fac = dC[:-1] * dC[1:] - C * Cinv_offdiag
    bcpj_fac = B * dC[:-1] - dB[:-1] * C

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        # best-fit a, b, c
        c = bcpj_fac / cpj_fac - (B * dC[1:] - dB[1:] * C) / cjck_fac
        c /= cjck_fac / cpj_fac - (dC[1:] ** 2 - C * Cinv_diag[1:]) / cjck_fac
        b = (bcpj_fac - c * cjck_fac) / cpj_fac
        a = (B - b * dC[:-1] - c * dC[1:]) / C
        result.countrate_two_omit = a

        # best-fit chi squared
        result.chisq_two_omit = (
            A + a**2 * C + b**2 * Cinv_diag[:-1] + c**2 * Cinv_diag[1:]
        )
        result.chisq_two_omit -= 2 * a * B + 2 * b * dB[:-1] + 2 * c * dB[1:]
        result.chisq_two_omit += (
            2 * a * b * dC[:-1] + 2 * a * c * dC[1:] + 2 * b * c * Cinv_offdiag
        )
        result.chisq_two_omit /= scale

        # uncertainty on the slope from inverting the (a, b, c)
        # covariance matrix
        fac = Cinv_diag[1:] * Cinv_diag[:-1] - Cinv_offdiag**2
        term2 = dC[:-1] * (dC[:-1] * Cinv_diag[1:] - Cinv_offdiag * dC[1:])
        term3 = dC[1:] * (dC[:-1] * Cinv_offdiag - Cinv_diag[:-1] * dC[1:])
        result.uncert_two_omit = np.sqrt(fac / (C * fac - term2 + term3))
        result.uncert_two_omit *= np.sqrt(scale)

    result.fill_masked_reads(diffs2use)

    return result


def compute_alphas_betas(count_rate_guess, gain, rnoise, covar, rescale, diffs, dn_scale):
    """
    Compute alpha, beta, and scale needed for ramp fit.

    Elements of the covariance matrix.

    Parameters
    ----------
    count_rate_guess : ndarray
        Initial guess (ncols,)

    gain : ndarray
        Gain (ncols,)

    rnoise : ndarray
        Read noise (ncols,)

    covar : Covar
        The class instance that computes and contains the covariance matrix info.

    rescale : bool
        Determination to rescale covariance matrix.

    diffs : ndarray
        The group differences of the data (ngroups-1, nrows, ncols).

    dn_scale : float
        Scaling factor.

    Returns
    -------
    alpha : ndarray
        Diagonal of covariance matrix.

    beta : ndarray
        Off diagonal of covariance matrix.

    scale : ndarray or integer
        Overflow/underflow prevention scale.

    """
    warnings.filterwarnings("ignore", ".*invalid value.*", RuntimeWarning)
    warnings.filterwarnings("ignore", ".*divide by zero.*", RuntimeWarning)

    alpha_phnoise = count_rate_guess / gain * covar.alpha_phnoise[:, np.newaxis]
    alpha_readnoise = rnoise**2 * covar.alpha_readnoise[:, np.newaxis]
    alpha = alpha_phnoise + alpha_readnoise

    beta_phnoise = count_rate_guess / gain * covar.beta_phnoise[:, np.newaxis]
    beta_readnoise = rnoise**2 * covar.beta_readnoise[:, np.newaxis]
    beta = beta_phnoise + beta_readnoise

    warnings.resetwarnings()

    ndiffs, npix = diffs.shape

    # Rescale the covariance matrix to a determinant of 1 to
    # avoid possible overflow/underflow.  The uncertainty and chi
    # squared value will need to be scaled back later.  Note that
    # theta[-1] is the determinant of the covariance matrix.
    #
    # The method below uses the fact that if all alpha and beta
    # are multiplied by f, theta[i] is multiplied by f**i.  Keep
    # a running track of these factors to construct the scale at
    # the end, and keep scaling throughout so that we never risk
    # overflow or underflow.

    if rescale:
        # scale = np.exp(np.mean(np.log(alpha), axis=0))
        theta = np.ones((ndiffs + 1, npix))
        theta[1] = alpha[0]

        scale = theta[0] * 1
        warnings.filterwarnings("ignore", ".*invalid value.*", RuntimeWarning)
        warnings.filterwarnings("ignore", ".*divide by zero.*", RuntimeWarning)

        for i in range(2, ndiffs + 1):
            theta[i] = (
                alpha[i - 1] / scale * theta[i - 1]
                - beta[i - 2] ** 2 / scale**2 * theta[i - 2]
            )

            # Scaling every ten steps in safe for alpha up to 1e20
            # or so and incurs a negligible computational cost for
            # the fractional power.

            if i % int(dn_scale) == 0 or i == ndiffs:
                f = theta[i] ** (1 / i)
                scale *= f
                tmp = theta[i] / f
                theta[i - 1] /= tmp
                theta[i - 2] /= tmp / f
                theta[i] = 1

        warnings.resetwarnings()
    else:
        scale = 1

    alpha /= scale
    beta /= scale

    alpha_tuple = (alpha, alpha_phnoise, alpha_readnoise)
    beta_tuple = (beta, beta_phnoise, beta_readnoise)

    return alpha_tuple, beta_tuple, scale


def compute_thetas(ndiffs, npix, alpha, beta):
    """
    Computes intermediate theta values for ramp fitting.

    EQNs 38-40

    Parameters
    ----------
    ndiffs : int
        Number of differences.

    npix : int
        Number of columns in a row.

    alpha : ndarray
        Diagonal of covariance matrix.

    beta : ndarray
        Off diagonal of covariance matrix.

    Returns
    -------
    theta : ndarray
    """
    theta = np.ones((ndiffs + 1, npix))
    theta[1] = alpha[0]
    for i in range(2, ndiffs + 1):
        theta[i] = alpha[i - 1] * theta[i - 1] - beta[i - 2] ** 2 * theta[i - 2]
    return theta


def compute_phis(ndiffs, npix, alpha, beta):
    """
    Computes intermediate phi values for ramp fitting.

    EQNs 41-43

    Parameters
    ----------
    ndiffs : int
        Number of differences.

    npix : int
        Number of columns in a row.

    alpha : ndarray
        Diagonal of covariance matrix.

    beta : ndarray
        Off diagonal of covariance matrix.

    Returns
    -------
    phi : ndarray
    """
    phi = np.ones((ndiffs + 1, npix))
    phi[ndiffs - 1] = alpha[ndiffs - 1]
    for i in range(ndiffs - 2, -1, -1):
        phi[i] = alpha[i] * phi[i + 1] - beta[i] ** 2 * phi[i + 2]
    return phi


def compute_Phis(ndiffs, npix, beta, phi, sgn):
    """
    Computes intermediate Phi values for ramp fitting.

    EQN 46

    Parameters
    ----------
    ndiffs : int
        Number of differences.

    npix : int
        Number of columns in a row.

    alpha : ndarray
        Diagonal of covariance matrix.

    beta : ndarray
        Off diagonal of covariance matrix.

    sgn : ndarray
        Oscillating 1, -1 sequence.

    Returns
    -------
    Phi : ndarray
    """
    Phi = np.zeros((ndiffs, npix))
    for i in range(ndiffs - 2, -1, -1):
        Phi[i] = Phi[i + 1] * beta[i] + sgn[i + 1] * beta[i] * phi[i + 2]
    return Phi


def compute_PhiDs(ndiffs, npix, beta, phi, sgn, diff_mask):
    """
    EQN 4, Paper II
    This one is defined later in the paper and is used for jump detection.

    Parameters
    ----------
    ndiffs : int
        Number of differences.

    npix : int
        Number of columns in a row.

    beta : ndarray
        Off diagonal of covariance matrix.

    phi : ndarray
        Intermediate computation.

    sgn : ndarray
        Oscillating 1, -1 sequence.

    diff_mask : ndarray
        Mask of differences used.

    Returns
    -------
    PhiD: ndarray
    """
    PhiD = np.zeros((ndiffs, npix))
    for i in range(ndiffs - 2, -1, -1):
        PhiD[i] = (PhiD[i + 1] + sgn[i + 1] * diff_mask[i + 1] * phi[i + 2]) * beta[i]
    return PhiD


def compute_Thetas(ndiffs, npix, beta, theta, sgn):
    """
    EQN 47

    Parameters
    ----------
    ndiffs : int
        Number of differences.

    npix : int
        Number of columns in a row.

    beta : ndarray
        Off diagonal of covariance matrix.

    theta : ndarray
        Intermediate computation.

    sgn : ndarray
        Oscillating 1, -1 sequence.

    Returns
    -------
    Theta : ndarray
    """
    Theta = np.zeros((ndiffs, npix))
    Theta[0] = -theta[0]
    for i in range(1, ndiffs):
        Theta[i] = Theta[i - 1] * beta[i - 1] + sgn[i] * theta[i]
    return Theta


def compute_ThetaDs(ndiffs, npix, beta, theta, sgn, diff_mask):
    """
    EQN 48

    Parameters
    ----------
    ndiffs : int
        Number of differences.

    npix : int
        Number of columns in a row.

    beta : ndarray
        Off diagonal of covariance matrix.

    theta : ndarray
        Intermediate computation.

    sgn : ndarray
        Oscillating 1, -1 sequence.

    diff_mask : ndarray
        Mask of differences used.

    Returns
    -------
    ThetaD : ndarray
    """
    ThetaD = np.zeros((ndiffs + 1, npix))
    ThetaD[1] = -diff_mask[0] * theta[0]
    for i in range(1, ndiffs):
        ThetaD[i + 1] = beta[i - 1] * ThetaD[i] + sgn[i] * diff_mask[i] * theta[i]
    return ThetaD


def matrix_computations(
    ndiffs, npix, sgn, diff_mask, diffs2use, beta, phi, Phi, PhiD, theta, Theta, ThetaD
):
    """
    Compute matrix computations needed for ramp fitting.

    EQNs 61-63, 71, 75

    Parameters
    ----------
    ndiffs : int
        Number of differences.

    npix : int
        Number of columns in a row.

    sgn : ndarray
        Oscillating 1, -1 sequence.

    diff_mask : ndarray
        Mask of differences used.

    diff2use : ndarray
        Masked differences.

    beta : ndarray
        Off diagonal of covariance matrix.

    phi : ndarray
        Intermediate computation.

    Phi : ndarray
        Intermediate computation.

    PhiD : ndarray
        Intermediate computation.

    theta : ndarray
        Intermediate computation.

    Theta : ndarray
        Intermediate computation.

    ThetaD : ndarray
        Intermediate computation.

    Returns
    -------
    dB : ndarray
        Intermediate computation.

    dC : ndarray
        Intermediate computation.

    A : ndarray
        Intermediate computation.

    B : ndarray
        Intermediate computation.

    C : ndarray
        Intermediate computation.
    """
    beta_extended = np.ones((ndiffs, npix))
    beta_extended[1:] = beta

    # C' and B' in the paper

    dC = sgn / theta[ndiffs] * (phi[1:] * Theta + theta[:-1] * Phi)
    dC *= diffs2use  # EQN 71

    dB = sgn / theta[ndiffs] * (phi[1:] * ThetaD[1:] + theta[:-1] * PhiD)  # EQN 75

    # {\cal A}, {\cal B}, {\cal C} in the paper

    # EQNs 61-63
    A = 2 * np.sum(
        diff_mask * sgn / theta[-1] * beta_extended * phi[1:] * ThetaD[:-1], axis=0
    )
    A += np.sum(diff_mask**2 * theta[:-1] * phi[1:] / theta[ndiffs], axis=0)

    B = np.sum(diff_mask * dC, axis=0)
    C = np.sum(dC, axis=0)

    return dB, dC, A, B, C


def get_ramp_result(
    dC, A, B, C, scale, alpha_phnoise, alpha_readnoise,
    beta_phnoise, beta_readnoise,
):
    """
    Use intermediate computations to fit the ramp and save the results.

    Parameters
    ----------
    dC : ndarray
        Intermediate computation.

    A : ndarray
        Intermediate computation.

    B : ndarray
        Intermediate computation.

    C : ndarray
        Intermediate computation.

    rescale : float
        Factor applied to each element of the covariance matrix to
        normalize its determinant in order to avoid possible
        overflow/underflow problems for long ramps.

    alpha_phnoise : ndarray
        The photon noise contribution to the alphas.

    alpha_readnoise : ndarray
        The read noise contribution to the alphas.

    beta_phnoise : ndarray
        The photon noise contribution to the betas.

    beta_readnoise : ndarray
        The read noise contribution to the betas.

    Returns
    -------
    result : RampResult
        The results of the ramp fitting for a given row of pixels in an integration.
    """
    result = RampResult()

    # Finally, save the best-fit count rate, chi squared, uncertainty
    # in the count rate, and the weights used to combine the
    # groups.

    warnings.filterwarnings("ignore", ".*invalid value.*", RuntimeWarning)
    warnings.filterwarnings("ignore", ".*divide by zero.*", RuntimeWarning)
    invC = 1 / C
    result.countrate = B * invC
    result.chisq = (A - B**2 / C) / scale

    result.uncert = np.sqrt(scale / C)
    result.weights = dC / C

    result.var_poisson = np.sum(result.weights**2 * alpha_phnoise, axis=0)
    result.var_poisson += 2 * np.sum(
        result.weights[1:] * result.weights[:-1] * beta_phnoise, axis=0
    )

    result.var_rnoise = np.sum(result.weights**2 * alpha_readnoise, axis=0)
    result.var_rnoise += 2 * np.sum(
        result.weights[1:] * result.weights[:-1] * beta_readnoise, axis=0
    )

    warnings.resetwarnings()

    return result


################################################################################
################################## DEBUG #######################################


def dbg_print_info(group_time, readtimes, data, diff):
    print(DELIM)
    print(f"group_time = {group_time}")
    print(DELIM)
    print(f"readtimes = {array_string(np.array(readtimes))}")
    print(DELIM)
    print(f"data = {array_string(data[:, 0, 0])}")
    print(DELIM)
    data_gt = data / group_time
    print(f"data / gt = {array_string(data_gt[:, 0, 0])}")
    print(DELIM)
    print(f"diff = {array_string(diff[:, 0, 0])}")
    print(DELIM)


def array_string(arr, prec=4):
    return np.array2string(arr, precision=prec, max_line_width=np.nan, separator=", ")
